# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import base64
import hmac
import hashlib
import logging
import os
import urllib.parse
import uuid

from majormode.perseus.constant.application import ApplicationPlatform
from majormode.perseus.constant.application import ApplicationStage
from majormode.perseus.constant.obj import ObjectStatus
from majormode.perseus.service.account.account_service import AccountService
from majormode.perseus.service.base_rdbms_service import BaseRdbmsService
from majormode.perseus.service.base_service import BaseService
from majormode.perseus.service.team.team_service import TeamService
from majormode.perseus.utils import cast

import settings


class ApplicationService(BaseRdbmsService):
    """

    """
    # Define the name of the Content Delivery Network (CDN) bucket that
    # groups application logos all together.
    CDN_BUCKET_NAME_PICTURE = 'app'

    # Hash algorithm to use to generate the Consumer Secret key of an
    # application.
    CRYPTOGRAPHIC_HASH_FUNCTION = hashlib.sha1

    # # Determine which properties passed to this function have changed.  If
    # # no property has changed, the function simply returns.
    # IS_PROPERTY_NULLABLE_DICT = {
    #     'app_name': False,
    #     'platform': False,
    #     'logo_url': True,
    # }

    class ApplicationNameAlreadyRegisteredException(BaseService.BaseServiceException):
        """
        Signal the a specified application name is already used by another
        client application registered against the platform.
        """

    class IncorrectSignatureException(BaseService.BaseServiceException):
        """
        Signal that signature of the HTTP request is incorrect.  The API
        Secret Key that has been used to sign the request might no correspond
        to the associated API Consumer Key.
        """

    def _assert_administrator(self, app_id, account_id):
        """
        Assert that the specified user is an administrator of the application,
        either as an individual, either as an administrator of the team that
        owns this application.


        :param app_id: Identification of a application to check whether the
            specified user is administrator of.

        :param account_id: Identification of the account of a user to check
            whether he has the role of administrator for the specified
            application.


        :raise IllegalAccessException: If the user specified by his account is
           not an administrator of the application.
        """
        application = self.get_application(app_id, check_status=True)
        if application.team_id is None:
            if account_id != application.account_id:
                raise self.IllegalAccessException()
        else:
            TeamService()._assert_administrator(account_id, application.team_id)

    @classmethod
    def build_picture_url(cls, picture_id):
        """
        Return the Uniform Resource Locator of the picture of an application.


        @param picture_id: The identification of the picture of an application.


        @return: A string representing the Uniform Resource Locator of the
            picture.
        """
        return picture_id and os.path.join(
            settings.CDN_URL_HOSTNAME,
            cls.CDN_BUCKET_NAME_PICTURE,
            str(picture_id))

    def get_application(
            self,
            app_id,
            check_status=False,
            connection=None):
        """
        Return extended information about the application specified by its
        identification.


        @warning: This function is for internal usage only; it MUST not be
            surfaced to client applications.


        :param app_id: Identification of the application that accesses the service.

        :param check_status: Indicate whether the function must check the
            current status of this application and raise an exception if it is
            not of enabled.

        :param connection: An object `RdbmsConnection`.


        :return: An object containing the following members:

            * ``account_id``: Identification of the account of the user who
              registered this client application.

            * ``app_id``: Identification of the client application.

            * ``consumer_key``: A unique string that identifies the API
              Consumer Key tied to the client application.

            * ``consumer_secret``: A string kept secret used to authenticate the API
              Consumer Key.

            * ``creation_time``: Time when the application has been registered.

            * ``name``: Name of the application.

            * ``object_status``: Current status of the application.

            * ``picture_id``: Identification of the logo of the application.

            * ``package``: Package name of the family this application belongs
              to, if any defined.  This allows to group under a same "branding
              umbrella" several applications that share the functionalities,
              the same users base. For instance, a software product can have a
              lite (free) and full (paid) versions on an Android and an iOS
              platforms, etc. Grouping applications of a same family allows to
              send notification to a user to whatever version(s) and platform(s)
              he could use.

            * ``platform``: Identification of the platform which the application
              runs on (cf. ``ApplicationPlatform``).

            * ``team_id``: Identification of the organization this application
              belongs to.

            * ``update_time``: Most recent time when some information of the
              application has been modified, such as its name or its API keys.


        :raise DeletedObjectException: If the application has been deleted,
            while the argument ``check_status`` has been set to ``True``.

        :raise DisabledObjectException: If the application has been disabled,
            while the argument ``check_status`` has been set to ``True``.

        :raise UndefinedObjectException: If the specified identification
            doesn't refer to a application registered against the platform.
        """
        # Retrieve the information of the application specified by its
        # identification.
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    app_id,
                    app_name
                    package_name,
                    picture_id,
                    stage,
                    platform,
                    consumer_key,
                    consumer_secret,
                    account_id,
                    team_id,
                    object_status,
                    creation_time,
                    update_time
                  FROM
                    application
                  WHERE
                    app_id = %(app_id)s
                """,
                {
                    'app_id': app_id
                }
            )
            row = cursor.fetch_one()

            if row is None:
                raise AccountService.UndefinedObjectException(f'Undefined application "{app_id}"')

            application = row.get_object({
                'account_id': cast.string_to_uuid,
                'app_id': cast.string_to_uuid,
                'creation_time': cast.string_to_timestamp,
                'object_status': ObjectStatus,
                'picture_id': cast.string_to_uuid,
                'platform': ApplicationPlatform,
                'stage': ApplicationStage,
                'team_id': cast.string_to_uuid,
                'update_time': cast.string_to_timestamp
            })

            # Check the current status of the application.
            if check_status:
                if application.object_status == ObjectStatus.deleted:
                    raise ApplicationService.DeletedObjectException(f'Application "{app_id}" is deleted')
                elif application.object_status == ObjectStatus.disabled:
                    raise ApplicationService.DisabledObjectException(f'Application "{app_id}" is disabled')

            return application

    def register_application(
            self,
            account_id,
            name,
            platform,
            package=None,
            picture_id=None,
            stage=ApplicationStage.sandbox,
            team_id=None):
        """
        Register an application that needs to access to the online service.

        Generate a new Application Programming Interface (API) key, which
        uniquely identifies this application.

        :param account_id: Identification of the account of the user who
           registers this application.  This account will own the API key that
           is issued.

        :param name: The name of the application.

        :param picture_id: Identification of the logo of the application.

        :param platform: An item of `ApplicationPlatform` that describes the
            platform that this application runs on.

        :param package: package name of the family this application belongs
            to, if any defined. This allows to group under a same "branding
            umbrella" several applications that share the functionalities, the
            same users base. For instance, a software product can have a lite
            (free) and full (paid) versions on an Android and an iOS platforms,
            etc. Grouping applications of a same family allows to send
            notification to a user to whatever version(s) and platform(s) he
            could use.

        :param stage: An item of ``ApplicationStage`` that indicates the
            environment stage which the application is deployed onto.

        :param team_id: Identification of a team this application belongs to.


        :return: An object that contains the following members:

            * ``app_id``: Identification of the application.

            * ``consumer_key``: A unique string that identifies the API Consumer
              Key.

            * ``consumer_secret``: A string that must be kept secret, which
              is used to authenticate the API Consumer Key.


            * ``picture_id``: Identification of the application's logo picture, if
              any defined.

            * ``picture_url``: Uniform Resource Locator (URL) that specifies the
              location of the application's logo picture, if any defined.  A
              client application can use this URL and append the query parameter
              ``size`` to specify a given pixel resolution of the picture, such as
              ``thumbnail``, ``small``, ``medium``, ``large``.

            * ``object_status``: Current status of the application.

            * ``creation_time``: Time when the application has been registered and
              the application API keys have been generated.

            * ``update_time``: Time of the most recent modification of one or more
              properties of this application.


        :raise ApplicationNameAlreadyRegisteredException: If the name of the
            application to register is already used for an other application.

        :raise DeletedObjectException: If the account of the user who registers
            this application has been deleted.

        :raise DisabledObjectException: If the account of the user who
            registers this application has been disabled.

        :raise UndefinedObjectException: If the argument ``account_id`` doesn't
            refer to a user account registered to the platform.
        """
        with self.acquire_rdbms_connection(True) as connection:
            # Retrieve information about the account of the user who registers this
            # application.
            AccountService().get_account(account_id, check_status=True, connection=connection)

            # Check whether there is no other application registered with the same
            # name for the same platform.
            cursor = connection.execute(
                """
                SELECT 
                    true
                  FROM 
                    application
                  WHERE 
                    lower(app_name) = lower(%(app_name)s)
                    AND platform = %(platform)s
                """,
                {
                    'app_name': name,
                    'platform': platform
                })

            if cursor.fetch_one() is not None:
                raise ApplicationService.ApplicationNameAlreadyRegisteredException('Application name already used')

            # Register the application and generate its API keys.
            cursor = connection.execute(
                """
                INSERT INTO application(
                    app_name,
                    platform,
                    stage,
                    package_name,
                    consumer_key,
                    consumer_secret,
                    picture_id,
                    account_id, 
                    team_id)
                  VALUES 
                    (%(app_name)s,
                     %(platform)s,
                     %(stage)s,
                     %(package_name)s,
                     %(consumer_key)s,
                     %(consumer_secret)s,
                     %(picture_id)s,
                     %(account_id)s,
                     %(team_id)s)
                  RETURNING 
                    app_id,
                    creation_time,
                    object_status,
                    consumer_key,
                    consumer_secret,
                    picture_id,
                    stage,
                    update_time
                """,
                {
                    'account_id': account_id,
                    'app_name': name,
                    'consumer_key': uuid.uuid1().hex,
                    'consumer_secret': base64.urlsafe_b64encode(uuid.uuid4().hex.encode()).decode(),
                    'package_name': package,
                    'picture_id': picture_id,
                    'platform': platform,
                    'stage': stage or ApplicationStage.sandbox,
                    'team_id': team_id
                })
            row = cursor.fetch_one()

            # Return the properties of this application.
            application = row.get_object({
                'app_id': cast.string_to_uuid,
                'creation_time': cast.string_to_timestamp,
                'object_status': ObjectStatus,
                'picture_id': cast.string_to_uuid,
                'stage': ApplicationStage,
                'update_time': cast.string_to_timestamp
            })

            application.picture_url = application.picture_id and os.path.join(
                settings.CDN_URL,
                self.CDN_BUCKET_NAME_PICTURE, str(application.picture_id))

            return application

    def validate_signature(
            self,
            consumer_key,
            path_query_string,
            message_body,
            api_sig=None,
            strict=True):
        """
        Check whether the Uniform Resource Locator (URL) specified in a HTTP
        request performed by a client application against the platform, is
        correctly signed with the secret key associated to the API key of the
        client application.


        :param consumer_key: Application Programming Interface (API) key tied
            to the application that accesses the service.

        :param path_query_string: Path and query string of the HTTP request
            that the application must have signed with its API secret key.

        :param message_body: Content of the HTTP request message body, or
            ``None`` if no message body has been passed along with the HTTP
            request.

        :param api_sig: Signature of the HTTP request as calculated by the
            client application.

        :param strict: Indicate whether the signature has to be verified or if
            the function should only checked that the application is at least
            enabled.  For instance, development environments should not require
            HTTP requests to be signed so that developers can easily debug them
            through command line tools such as cURL.


        :return: The identification of the application.


        :raise DeletedObjectException: If the specified API key has been
            deleted, and as such, it cannot be used or referenced anymore.

        :raise DisabledObjectException: If the specified API key has been
            disabled, and as such, it cannot be used anymore

        :raise IncorrectSignatureException: If the HTTP request has not been
            signed while the signature is required (production stage) or
            requested (argument ``strict``), or if the signature of the HTTP
            request is incorrect.  The API Secret Key that has been used to
            sign the request might no correspond to the associated API
            Consumer Key.

        :raise UndefinedObjectException: If the specified API key has not been
            registered against the platform.
        """
        # Retrieve information about the application.
        with self.acquire_rdbms_connection() as connection:
            cursor = connection.execute(
                """
                SELECT 
                    app_id,
                    consumer_secret,
                    stage,
                    object_status
                  FROM
                    application
                  WHERE
                    consumer_key = %(consumer_key)s
                """,
                {
                    'consumer_key': consumer_key
                }
            )
            row = cursor.fetch_one()
            if row is None:
                raise self.UndefinedObjectException(f"Undefined API consumer key {consumer_key}")

            application = row.get_object({
                'app_id': cast.string_to_uuid,
                'object_status': ObjectStatus,
                'stage': ApplicationStage})

        # Check the current status of the application.
        if application.object_status == ObjectStatus.deleted:
            raise ApplicationService.DeletedObjectException(f"The API consumer key {consumer_key} has been deleted")
        elif application.object_status == ObjectStatus.disabled:
            raise ApplicationService.DisabledObjectException(f"The API consumer key {consumer_key} has been disabled")

        # Check whether the HTTP request signature is required.
        if (application.stage == ApplicationStage.live or strict) and api_sig is None:
            raise self.IncorrectSignatureException('Undefined HTTP request signature')

        # Check whether the HTTP request signature is valid.
        if api_sig:
            # Replace %xx escapes from the query string with their single-character
            # equivalent.  Client applications generally build query string without
            # escaping special characters, while the HTTP library they used will do
            # encode them behind the scene.
            #
            # For example, one client application prepare and sign the following URL:
            #
            # ```url
            # /school/child?limit=100&offset=0&sync_time=2020-10-15T09:55:52.108+07:00'
            # ```
            #
            # But, its HTTP library encodes and sends to the RESTful API server the
            # following URL:
            #
            # ```url
            # /school/child?limit=100&offset=0&sync_time=2020-10-15T09%3A55%3A52.108%2B07%3A00"
            # ```
            message = urllib.parse.unquote(path_query_string)

            # Add the decoded version of the message body, if any.
            if message_body is not None:
                message += message_body.decode()

            digest = hmac \
                .new(
                    application.consumer_secret.encode(),
                    msg=message.encode(),
                    digestmod=ApplicationService.CRYPTOGRAPHIC_HASH_FUNCTION) \
                .hexdigest()

            if api_sig.lower() != digest:
                logging.debug(f'Path query string: "{path_query_string}"')
                logging.debug(f'Message body: "{"" if message_body is None else message_body.decode()}"')
                logging.debug(f'Encoded message: "{message}"')
                logging.debug(f'Expected signature: "{digest}"')
                logging.debug(f'Client signature: "{api_sig.lower()}"')
                raise ApplicationService.IncorrectSignatureException("Invalid HTTP request signature")

        return application.app_id

    # def update_application(self, app_id, account_id, this_app_id, properties):
    #     """
    #     Update some modifiable properties of the specified client application.
    #
    #
    #     :param app_id: identification of the client application such as a Web,
    #         a desktop, or a mobile application, that accesses the service.
    #
    #     :param account_id: identification of the account of the user who is
    #         updating the role of the member.  This first user MUST have a role
    #         of administrator.
    #
    #     :param this_app_id: identification of the client application for which
    #         some properties are updated.
    #
    #     :param properties: a dictionary of properties of the specified client
    #         application to be modified.
    #
    #
    #     :raise DeletedObjectException: if the client application has been
    #         deleted.
    #
    #     :raise DisabledObjectException: if the client application has been
    #         disabled.
    #
    #     :raise IllegalAccessException: if the user on behalf of some
    #         properties of the client application are modified is not the agent
    #         or one of the regular administrators of this application.
    #
    #     :raise UndefinedObjectException: if the specified identification
    #         doesn't refer to a client application registered to the platform.
    #     """
    #     with self.acquire_rdbms_connection(True) as connection:
    #         # Retrieve information about the account that registers this API key,
    #         # and check whether the current status of this account allows him to
    #         # register an API key.
    #         AccountService().get_account(account_id, check_status=True)
    #
    #         application = self.get_application(this_app_id, check_status=True)
    #
    #         if account_id != application.account_id:
    #             cursor = connection.execute("""
    #                 SELECT member_role
    #                   FROM application_member
    #                   WHERE app_id = %(app_id)s
    #                     AND account_id = %(account_id)s""",
    #                 { 'account_id': account_id,
    #                   'app_id': this_app_id })
    #             row = cursor.fetch_one()
    #             if row is None or row.get_value('member_role') != str(ApplicationService.MemberRole.administrator):
    #                 raise ApplicationService.IllegalAccessException('Not an administrator of the client application')
    #
    #         # Check whether there is no other API key registered with the same
    #         # name.
    #         name = properties.get('name')
    #         if name is not None and name != application.name:
    #             cursor = connection.execute("""
    #                 SELECT true
    #                   FROM application
    #                   WHERE lower(app_name) = lower(%(app_name)s)""",
    #                 { 'app_name': name })
    #             if cursor.get_row_count() > 0:
    #                 raise ApplicationService.ApplicationNameAlreadyRegisteredException('Application name already registered')
    #
    #         modified_properties = dict(
    #             [ (name, value) for (name, value) in properties.iteritems()
    #                 if ((self.IS_PROPERTY_NULLABLE_DICT[name] or value is not None) and \
    #                     getattr(application, name) != value) ])
    #
    #         if modified_properties:
    #             update_statement = [ "%s = %%(%s)s" % (name, name) for (name, value) in modified_properties.items() ]
    #             modified_properties['app_id'] = this_app_id
    #
    #             cursor = connection.execute("""
    #                 UPDATE application
    #                   SET """ + ','.join(update_statement) + """,
    #                       update_time = current_timestamp
    #                   WHERE app_id = %(app_id)s""",
    #               modified_properties)

    # def get_application_by_name(self, name, check_status=False):
    #     """
    #     Return extended information about the client application specified by
    #     its name.
    #
    #
    #     @warning: this function is for internal usage only; it MUST not be
    #         surfaced to client applications.
    #
    #
    #     :param name: name of the application to return information.
    #
    #     :param check_status: indicate whether the function must check the
    #         current status of this user account and raise an exception if it is
    #         not of enabled.
    #
    #     :return: an instance containing the following members:
    #
    #         * ``account_id``: identification of the account of the user who
    #           registered this client application.
    #
    #         * ``app_id``: identification of the application.
    #
    #         * ``app_name``: the name of the application.
    #
    #         * ``consumer_key``: a unique string that identifies the API key
    #           tied to the client application.
    #
    #         * ``consumer_secret``: a string kept secret, which is used to
    #           authenticate the API key.
    #
    #         * ``creation_time``: time when the application has been registered.
    #
    #         * ``picture_id``: identification of the logo of the client
    #           application.
    #
    #         * ``object_status``: current status of the application.
    #
    #         * ``package_name``: package name of the family this application belongs
    #           to, if any defined. This allows to group under a same "branding
    #           umbrella" several applications that share the functionalities,
    #           the same users base. For instance, a software product can have a
    #           lite (free) and full (paid) versions on an Android and an iOS
    #           platforms, etc. Grouping applications of a same family allows to
    #           send notification to a user to whatever version(s) and platform(s)
    #           he could use.
    #
    #         * ``platform``: identification of the platform on which the client
    #          application runs (cf. ``ApplicationService.ApplicationPlatform``).
    #
    #         * ``team_id``: identification of the team this client application
    #           belongs to.
    #
    #         * ``update_time``: most recent time when some information, such as
    #           the name of the application, or the API keys, has been modified.
    #
    #
    #     :raise DeletedObjectException: if the client application has been
    #        deleted, while the argument ``check_status`` has been set to
    #        ``True``.
    #
    #     :raise DisabledObjectException: if the client application has been
    #        disabled, while the argument ``check_status`` has been set to
    #        ``True``.
    #
    #     :raise UndefinedObjectException: if the specified name doesn't refer
    #        to a client application registered against the platform.
    #     """
    #     with self.acquire_rdbms_connection() as connection:
    #         cursor = connection.execute("""
    #             SELECT app_id,
    #                    app_name,
    #                    package_name,
    #                    picture_id,
    #                    platform,
    #                    consumer_key,
    #                    consumer_secret,
    #                    account_id,
    #                    team_id,
    #                    object_status,
    #                    creation_time,
    #                    update_time
    #               FROM application
    #               WHERE lower(app_name) = %(name)s""",
    #             { 'name': name.strip().lower() })
    #         row = cursor.fetch_one()
    #         if row is None:
    #             raise AccountService.UndefinedObjectException('The client application "%s" is not registered against the platform' % name)
    #         application = row.get_object({
    #                 'account_id': cast.string_to_uuid,
    #                 'app_id': cast.string_to_uuid })
    #
    #         if check_status:
    #             if application.object_status == ObjectStatus.disabled:
    #                 raise ApplicationService.DisabledObjectException('The client application referenced by the specified identification has been deleted')
    #
    #             if application.object_status == ObjectStatus.deleted:
    #                 raise ApplicationService.DeletedObjectException('The client application reference by the specified API key has been disabled')
    #
    #         return application

    # def get_applications(self, app_id, account_id):
    #     """
    #     Return extended information about client applications that the end
    #     user on behalf of this function is called has access to, either as a
    #     developer, either as an administrator.
    #
    #
    #     :param app_id: identification of the client application such as a Web,
    #         a desktop, or a mobile application, that accesses the service.
    #
    #     :param account_id: identification of a user account.
    #
    #
    #     :return: a list of instances containing the following members:
    #
    #         * ``account_id``: identification of the account of the user who
    #           registered this client application.
    #
    #         * ``app_id``: identification of the application.
    #
    #         * ``app_name``: the title of the application.
    #
    #         * ``consumer_key``: a unique string that identifies the API key
    #           tied to the client application.
    #
    #         * ``consumer_secret``: a string kept secret, which is used to
    #           authenticate the API key.
    #
    #         * ``creation_time``: time when the application has been registered.
    #
    #         * ``picture_id``: identification of the logo of the client
    #           application.
    #
    #         * ``object_status``: current status of the application.
    #
    #         * ``package_name``: package name of the family this application belongs
    #           to, if any defined. This allows to group under a same "branding
    #           umbrella" several applications that share the functionalities,
    #           the same users base. For instance, a software product can have a
    #           lite (free) and full (paid) versions on an Android and an iOS
    #           platforms, etc. Grouping applications of a same family allows to
    #           send notification to a user to whatever version(s) and platform(s)
    #           he could use.
    #
    #         * ``platform``: identification of the platform on which the client
    #          application runs (cf. ``ApplicationService.ApplicationPlatform``).
    #
    #         * ``team_id``: identification of the team this client application
    #           belongs to.
    #
    #         * ``update_time``: most recent time when some information, such as
    #           the name of the application, or the API keys, has been modified.
    #
    #
    #     :raise DeletedObjectException: if the user account has been deleted.
    #
    #     :raise DisabledObjectException: if the user account has been disabled.
    #
    #     :raise UndefinedObjectException: if the specified identification
    #         doesn't refer to a user account registered against the platform.
    #     """
    #     AccountService().get_account(account_id, check_status=True)
    #
    #     with self.acquire_rdbms_connection() as connection:
    #         # Retrieve the list of identification of the applications that the
    #         # specified user has access to, either as an individual, either as
    #         # the member of a team.
    #         teams = dict([ (team.team_id, team) for team in
    #               TeamService()._get_all_teams(app_id, account_id) ])
    #
    #         cursor = connection.execute("""
    #             SELECT app_id,
    #                    app_name,
    #                    package_name,
    #                    picture_id,
    #                    platform,
    #                    consumer_key,
    #                    consumer_secret,
    #                    account_id,
    #                    team_id,
    #                    object_status,
    #                    creation_time,
    #                    update_time
    #               FROM application
    #               WHERE app_id IN (
    #                 SELECT app_id
    #                   FROM application
    #                   WHERE account_id = %(account_id)s
    #                     AND team_id IS NULL
    #                 UNION ALL
    #                 SELECT app_id
    #                   FROM application
    #                   WHERE team_id IN (%[team_ids]s))""",
    #             { 'account_id': account_id,
    #               'team_ids': [ team.team_id for team in teams.iterkeys() ] })
    #         applications = [ row.get_object({
    #               'account_id': cast.string_to_uuid,
    #               'app_id': cast.string_to_uuid,
    #               'creation_time': cast.string_to_timestamp,
    #               'update_time': cast.string_to_timestamp })
    #             for row in cursor.fetch_all() ]
    #
    #         for application in applications:
    #             if application.team_id is not None:
    #                 application.team = team[application.team_id]
    #
    #         return applications.values()



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import json
import jsonpickle
import time

from majormode.perseus.model.obj import Object
from majormode.perseus.service.base_rdbms_service import BaseRdbmsService

import settings


class StatusService(BaseRdbmsService):
    def check_status(self):
        """
        Get the status of the different software components that the platform
        is built on.

        @return: a dictionary of software component status, where the key
                 corresponds to the name of the component and the value
                 contains the following attributes:
                 * ``name``: name of the software component.
                 * ``status``: boolean value that indicates whether the
                   component is available.
                 * ``check_duration``: duration, expressed in milliseconds, of
                   the check.
                 * ``error``: possible error detected by the method
                   responsible for checking this component.
        """
        class ComponentStatus(object):
            def __init__(self, name, status, check_duration, error=None):
                self.name = name
                self.status = status
                self.check_duration = check_duration
                if error is not None:
                    self.error = error

        def check_amqp():
            MESSAGE_TYPE_ON_LBS_STATUS_CHECKED = 'on_lbs_status_checked'
            with self.acquire_amqp_connection() as connection:
                connection.publish(MESSAGE_TYPE_ON_LBS_STATUS_CHECKED,
                    { 'check_time': time.strftime('%Y-%m-%dT%H:%M:%S.000+00', time.gmtime()) })

        def check_memcached():
            """
            Check whether memcached is up and running.

            The function tries to read the value of a key from memcached and to
            write a new value for this key corresponding to the current time.
            """
            import pylibmc

            MEMCACHED_CHECK_STATUS_KEY = 'cerberus.status.last_check_time'

            memcached = pylibmc.Client(['%s:%d' % (settings.MEMCACHED_SERVER_HOSTNAME, settings.MEMCACHED_SERVER_PORT) ], binary=True)
            memcached.behaviors = {"tcp_nodelay": True, "ketama": True}
            memcached.get(MEMCACHED_CHECK_STATUS_KEY)

            status = Object()
            status.last_check_time = time.strftime('%Y-%m-%dT%H:%M:%S.000+00', time.gmtime())

            memcached.set(MEMCACHED_CHECK_STATUS_KEY,
                          json.dumps(jsonpickle.Pickler(unpicklable=False).flatten(status)))

        def check_postgresql():
            """
            Check whether PostgreSQL is up and running.

            The function checks whether the status table has been created.  If
            not, the function creates it and stored the current time, otherwise it
            deletes any content from this table and stored the current time.
            """
            POSTGRESQL_CHECK_TABLE_NAME = '_cerberus_status_'

            with self.acquire_rdbms_connection() as connection:
                cursor = connection.execute("""
                    SELECT a.attname as "name",
                           pg_catalog.format_type(a.atttypid, a.atttypmod) as "data_type"
                       FROM pg_catalog.pg_attribute a
                       WHERE a.attnum > 0
                         AND NOT a.attisdropped
                         AND a.attrelid = (
                           SELECT c.oid
                             FROM pg_catalog.pg_class c
                             LEFT JOIN pg_catalog.pg_namespace n
                               ON n.oid = c.relnamespace
                             WHERE c.relname ~ '^(%s)$'
                             AND pg_catalog.pg_table_is_visible(c.oid))""" % POSTGRESQL_CHECK_TABLE_NAME)
                columns = [ row.get_object() for row in cursor.fetch_all() ]
                for column in columns:
                    if column.name == 'last_check_time' and \
                       column.data_type == 'timestamp with time zone':
                        connection.execute("""
                            DELETE FROM %s
                              RETURNING last_check_time""" % POSTGRESQL_CHECK_TABLE_NAME)
                        connection.execute("""
                            INSERT INTO %s(last_check_time)
                              VALUES (current_timestamp)""" % POSTGRESQL_CHECK_TABLE_NAME)
                        break
                else:
                    if len(columns) > 0:
                        connection.execute('DROP TABLE %s' % POSTGRESQL_CHECK_TABLE_NAME)
                    connection.execute("""
                        CREATE TABLE %s(
                          last_check_time timestamp with time zone NOT NULL DEFAULT current_timestamp)""" % POSTGRESQL_CHECK_TABLE_NAME)

                connection.commit()

        def check_redis():
            """
            Check whether redis is up and running.

            The function tries to read the value of a key from redis and to
            write a new value for this key corresponding to the current time.
            """
            import redis

            REDIS_CHECK_STATUS_KEY = 'cerberus.status.last_check_time'

            status = Object()
            status.last_check_time = time.strftime('%Y-%m-%dT%H:%M:%S.000+00', time.gmtime())

            r = redis.StrictRedis(host='localhost', port=6379, db=0)
            r.get(REDIS_CHECK_STATUS_KEY)
            r.set(REDIS_CHECK_STATUS_KEY,
                  json.dumps(jsonpickle.Pickler(unpicklable=False).flatten(status)))

        CHECK_METHODS = [
            ('memcached', check_memcached),
            ('postgresql', check_postgresql),
            ('redis', check_redis)
            #('amqp', check_amqp),
        ]

        component_statuses = []
        for (component_name, check_method) in CHECK_METHODS:
            try:
                begin_time = datetime.datetime.now()
                check_method()
                end_time = datetime.datetime.now()
                component_statuses.append(ComponentStatus(component_name, True, end_time - begin_time))
            except Exception as exception:
                end_time = datetime.datetime.now()
                component_statuses.append(
                    ComponentStatus(component_name, False, end_time - begin_time, error=repr(exception)))

        return component_statuses



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with
# the terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE
# SHALL NOT BE LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE
# AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
# DERIVATIVES.
#
# @version $Revision$

import argparse
import re

REGEX_GEOLITE_COUNTRY = re.compile(r'^"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}","\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}","(\d+)","(\d+)","([A-Z]{2})",".*"$')

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='GeoLite Country Importer')
    parser.add_argument('--file')
    arguments = parser.parse_args()

    with open(arguments.file) as file:
        print 'INSERT INTO _ip2country(from_ip, to_ip, country_code) VALUES %s' % \
            ','.join([ "(%s,%s,'%s')" % (match.group(1), match.group(2), match.group(3))
                         for match in [ REGEX_GEOLITE_COUNTRY.match(line) for line in file.readlines() ] if match ])



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import json
import socket
import urllib.request

from majormode.perseus.constant.http import HttpMethod
from majormode.perseus.constant.notification import DevicePlatform
from majormode.perseus.constant.obj import ObjectStatus
from majormode.perseus.constant.sort_order import SortOrder
from majormode.perseus.model import obj
from majormode.perseus.model.locale import Locale
from majormode.perseus.utils import cast
from majormode.perseus.utils.date_util import ISO8601DateTime

from majormode.perseus.service.application.application_service import ApplicationService
from majormode.perseus.service.base_rdbms_service import BaseRdbmsService
from majormode.perseus.service.notification.gcm import GoogleCloudMessagingRequest
from majormode.perseus.service.team.team_service import TeamService
import settings


class NotificationService(BaseRdbmsService):
    """
    A notification is a lightweight message that needs to be delivered to
    one or more recipients.  It informs the recipients about an event
    that occurs, which might require fetching additional data from the
    server platform or requesting the recipient to perform some action.

    A recipient is generally a client application that acts on behalf of a
    user, but it could also be an agent or botnet that controls a device.

    A message can be delivered to a recipient using different styles of
    network communication:

    * `email`: the message is delivered in an Internet electronic mail
      message to the specified recipients based on a store-and-forward
      model.

    * `push`: the message is delivered to the specified recipients
      when the request for the transmission of information is initiated
      by the publisher or server platform and pushed out to the receiver
      or client application.

    * `pull`: the message is delivered to the specified recipients
      when the request for the transmission of information is initiated
      by the receiver or client application, and then is responded by the
      publisher or server platform.  Push style requires recipients to
      register with the server platform before it can receive messages
      using this mode.
    """

    # Default and maximum durations, expressed in seconds, during which
    # information of the notification reasonably may be expected usefully
    # to inform an action or interest of the intended recipient.  After
    # the lifespan expires, provision of the notification to the intended
    # recipients may be prevented.
    DEFAULT_LIFESPAN = 24 * 60 * 60  # 1 day
    MAXIMUM_LIFESPAN = 4 * 7 * 24 * 60 * 60  # 4 weeks

    # Define the host name of the Google Cloud Messaging for Android, a
    # free service that helps developers send data from servers to their
    # Android applications on Android devices, and upstream messages from
    # the user's device back to the cloud.  This could be a lightweight
    # message telling the Android application that there is new data to be
    # fetched from the server (for instance, a "new email" notification
    # informing the application that it is out of sync with the back end),
    # or it could be a message containing up to 4kb of payload data (so
    # apps like instant messaging can consume the message directly).  The
    # GCM service handles all aspects of queueing of messages and delivery
    # to the target Android application running on the target device.
    GCM_SERVER_HTTP_URL = 'https://gcm-http.googleapis.com/gcm/send'

    def __add_device_registration(
            self,
            device_id,
            device_platform,
            app_id,
            device_token,
            account_id=None,
            connection=None,
            locale=None,
            utc_offset=None):
        """
        Register a mobile device to the push notification service


        :param device_id: Identification of a mobile device that registers
            to push notification service.

        :param device_platform: An item of the enumeration `DevicePlatform`.

        :param app_id: Identification of the mobile application that
            registers the mobile device to push notification service.

        :param device_token: Token that identifies the device by the push
            notification provider of the device platform.

        :param account_id: Identification of the account of a user who
            registers his mobile device to push notification service.

        :param connection: An object `RdbmsConnection`.

        :param locale: Preferred language to new messages' textual content in.
            This argument is ignored if the mobile device is registered to
            push notification on behalf of a user. Messages will be send in
            the current preferred language of the user.

        :param utc_offset: Difference between the location of the device and
            UTC (Universal Time Coordinated).


        :return: An object containing the following attributes:

            * `device_id` (required): Identification of the mobile device that
              registers to push notification service.

            * `object_status` (required): Current status of the registration of
              this device to push notification service.

            * `registration_id` (required): Identification of the registration of
              the mobile device to the push notification service.

            * `update_time` (required): Time of the most recent modification of
              information of the mobile device's registration to the push
              notification service.
        """
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            cursor = connection.execute("""
                INSERT INTO notification_device (
                    device_token,
                    device_id,
                    device_platform,
                    account_id,
                    utc_offset,
                    locale,
                    app_id)
                  VALUES (
                    %(device_token)s,
                    %(device_id)s,
                    %(device_platform)s,
                    %(account_id)s,
                    %(utc_offset)s,
                    %(locale)s,
                    %(app_id)s
                  )
                  RETURNING
                    device_id,
                    object_status,
                    registration_id,
                    update_time
                """,
                {
                    'account_id': account_id,
                    'app_id': app_id,
                    'device_id': device_id,
                    'device_platform': device_platform,
                    'device_token': device_token,
                    'locale': None if account_id else locale,
                    'utc_offset': utc_offset
                }
            )

            row = cursor.fetch_one()
            device_registration = row.get_object({
                'object_status': ObjectStatus,
                'registration_id': cast.string_to_uuid,
                'update_time': cast.string_to_timestamp,
            })

            return device_registration

    def __get_device_registration(
            self,
            device_id,
            app_id,
            account_id=None,
            connection=None):
        """
        Return the registration of a mobile device to push notification
        service


        :param device_id: Identification of a mobile device that registered
            to push notification service.

        :param app_id: Identification of the mobile application that
            registered the mobile device to push notification service.

        :param account_id: Identification of the account of a user who
            registered his mobile device to push notification service.

        :param connection: An object `RdbmsConnection`.


        :return: An object containing the following attributes:

            * `account_id` (optional): Identification of the account of the user
              who registered his mobile device to push notification service.

            * `app_id` (required): Identification of the mobile application that
              registered the mobile device to push notification service.

            * `device_id` (required): Identification of the mobile device that
              registered to push notification service.

            * `device_platform` (required): An item of the enumeration
              `DevicePlatform` describing the platform of the mobile device.

            * `device_token` (required): Token that identifies the mobile device
              by the push notification provider of the device platform.

            * `locale` (optional): Preferred language to receive new message's
              textual content in.  This argument is not used if the device
              registered to push notification service on behalf of a user, but the
              preferred language of this user.

            * `object_status` (required): Current status of the registration of
              this device to push notification service

            * `registration_id` (required): Identification of the registration of
              the mobile device to the push notification service.

            * `update_time` (required): Time of the most recent modification of
              information of the mobile device's registration to the push
              notification service.

            * `utc_offset` (optional): Difference between the location of the
               mobile device and UTC (Universal Time Coordinated).
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT
                    account_id,
                    app_id,
                    device_id,
                    device_platform,
                    device_token,
                    locale,
                    object_status,
                    registration_id,
                    update_time,
                    utc_offset
                  FROM 
                    notification_device
                  WHERE
                    device_id = %(device_id)s
                    AND ((account_id IS NULL AND %(account_id)s IS NULL)
                         OR account_id = %(account_id)s)
                    AND app_id = %(app_id)s
                """,
                {
                    'account_id': account_id,
                    'app_id': app_id,
                    'device_id': device_id
                })

            row = cursor.fetch_one()

            device_registration = row and row.get_object({
                'account_id': cast.string_to_uuid,
                'app_id': cast.string_to_uuid,
                'device_platform': DevicePlatform,
                'locale': cast.string_to_locale,
                'object_status': ObjectStatus,
                'registration_id': cast.string_to_uuid,
                'update_time': cast.string_to_timestamp,
            })

            return device_registration

    def __update_device_registration_token(
            self,
            registration_id,
            device_token,
            connection=None):
        """
        Update the token of a device registered to push notification service


        :param registration_id: Identification of the registration of a device
            to push notification service

        :param device_token: Token that identifies the device by the push
            notification provider of the device platform.

        :param connection: An object `RdbmsConnection` with auto commit.


        :return: An object containing the following attributes:

            * `device_id` (required): Identification of the mobile device that
              registers to push notification service.

            * `object_status` (required): Current status of the registration of
              this device to push notification service.

            * `registration_id` (required): Identification of the registration of
              the mobile device to the push notification service.

            * `update_time` (required): Time of the most recent modification of
              information of the mobile device's registration to the push
              notification service.
        """
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            cursor = connection.execute(
                """
                UPDATE 
                    notification_device
                  SET
                    device_token = %(device_token)s,
                    update_time = current_timestamp
                  WHERE
                    registration_id = %(registration_id)s
                  RETURNING
                    device_id,
                    object_status,
                    registration_id,
                    update_time
                """,
                {
                    'registration_id': registration_id,
                    'device_token': device_token
                })

            row = cursor.fetch_one()
            device_registration = row.get_object({
                'object_status': ObjectStatus,
                'registration_id': cast.string_to_uuid,
                'update_time': cast.string_to_timestamp,
            })

            return device_registration

    def _push_android_notification(
            self,
            google_api_key,
            device_tokens,
            notification,
            delay_while_idle=None,
            gcm_server_http_url=GCM_SERVER_HTTP_URL,
            time_to_live=None):
        """
        Push a notification to an Android device using Google Firebase Cloud
        Messaging (FCM) HTTP connection server, and return the result returned
        by the FCM connection server.


        :note: If your organization has a firewall that restricts the traffic
            to or from the Internet, you need to configure it to allow
            connectivity with GCM in order for your GCM client apps to receive
            messages. The ports to open are: 5228, 5229, and 5230. GCM
            typically only uses 5228, but it sometimes uses 5229 and 5230. GCM
            doesn't provide specific IPs, so you should allow your firewall to
            accept outgoing connections to all IP addresses contained in the
            IP blocks listed in Google's ASN of 15169.


        :param google_api_key: API key saved that gives the application server
            authorized access to Google services.

        :param device_tokens: a list of devices (registration tokens, or IDs as
            issued by the GCM connection servers to the client application)
             receiving a multicast message. It must contain at least 1 and at
             most 1000 registration IDs.

        :param notification: dictionary of key-value pairs of the notification.
            The key should not be a reserved word ("from" or any word starting
            with "google" or "gcm"). Values in string types are recommended.
            You have to convert values in objects or other non-string data
            types (e.g., integers or booleans) to string.

        :param delay_while_idle: indicate whether the notification should not
            be sent until the device becomes active.

        :param gcm_server_http_url: a specific URL to send the message to GCM
            connection server.

        :param time_to_live: specify how long in seconds the message should be
            kept in GCM storage if the device is offline.  The maximum time to
            live supported is 4 weeks.  The default value is 4 weeks.


        :return: `None` if the JSON request failed, otherwise a JSON object
            that contains the Downstream HTTP message response body:

            * `multicast_id:integer` (required): unique ID (number) identifying
              the multicast message.

            * `success:integer` (required): number of messages that were
              processed without an error.

            * `failure:integer` (required): number of messages that could not be
              processed.

            * `canonical_ids:integer` (required): number of results that contain
              a canonical registration token.

            * `results:list` (optional): array of objects representing the status
              of the messages processed. The objects are listed in the same order
              as the request (i.e., for each registration ID in the request, its
              result is listed in the same index in the response):

              * `message_id:string`: specify a unique ID for each successfully
                processed message.

              * `registration_id:string` (optional): specify the canonical
                registration token for the client app that the message was processed
                and sent to. Sender should use this value as the registration token
                for future requests. Otherwise, the messages might be rejected.

              * `error:string`: specify the error that occurred when processing
                the message for the recipient. The possible values can be found in
                `table 11 <https://developers.google.com/cloud-messaging/server-ref#table11>`_.
        """
        is_multicast_message = isinstance(device_tokens, (list, set, tuple))

        post_parameters = dict()
        post_parameters['registration_ids' if is_multicast_message else 'to'] = device_tokens

        if delay_while_idle:
            post_parameters['delay_while_idle'] = delay_while_idle

        post_parameters['priority'] = 10
        post_parameters['data'] = json.loads(obj.json_stringify(notification))

        if time_to_live:
            post_parameters['time_to_live'] = time_to_live

        headers = dict()
        headers['Authorization'] = 'key=%s' % google_api_key
        headers['Content-Type'] = 'application/json'

        try:
            request = GoogleCloudMessagingRequest(
                gcm_server_http_url,
                data=obj.jsonify(post_parameters),
                headers=headers,
                http_method=HttpMethod.POST)

            response = urllib.request.urlopen(request, timeout=None)
            data = response.read()

        except urllib.request.HTTPError as error:
            raise error

        except urllib.request.URLError as error:
            if isinstance(error.reason, socket.timeout):
                raise self.SocketTimeoutException()
            raise error

        except socket.timeout:
            raise self.SocketTimeoutException()

        response = data or json.loads(data)

        return response

    def get_notification(self, notification_id):
        """
        Return the properties of the specified notification.


        :param notification_id: identification of the notification to return
            the properties.


        :return: An object containing the following members:

            * `creation_time` (required): time when the sender originated
              the notification to the intended recipient.

            * `is_read` (required): indicate whether the notification
               has been read by the intended recipient.

            * `notification_id` (required): identification of the
              notification.

            * `notification_type` (required): string representation of the
              type of the notification, as selected by the sender that
              originated this notification to the intended recipient.

            * `payload` (optional): an arbitrary JSON expression added by
              the sender to provide information about the context of this
              notification.

            * `schedule_time` (required): time when this notification is
              scheduled to be sent to the intended recipient.
              The notification is not visible to the intended recipient prior
              to this time.

            * `sender_id` (optional): the identification of the sender that
              originated the notification.

            * `update_time` (required): time of the most recent modification
              of an attribute of the notification, such as its read status.
        """
        with self.acquire_rdbms_connection() as connection:
            cursor = connection.execute(
                """
                SELECT 
                    notification_id,
                    notification_type,
                    is_read,
                    sender_id,
                    payload,
                    schedule_time,
                    creation_time,
                    update_time
                  FROM
                    notification
                  WHERE
                    notification_id = %(notification_id)s
                """,
                {
                    'notification_id': notification_id
                })

            row = cursor.fetch_one()
            if row is None:
                raise self.UndefinedObjectException(f'undefined notification "{notification_id}"')

            notification = row.get_object({
                'creation_time': cast.string_to_timestamp,
                'notification_id': cast.string_to_uuid,
                'payload': cast.string_to_json,
                'schedule_time': cast.string_to_timestamp,
                'update_time': cast.string_to_timestamp
            })

            return notification

    def get_notifications(
            self,
            app_id,
            recipient_id,
            connection=None,
            end_time=None,
            include_read=False,
            limit=BaseRdbmsService.DEFAULT_LIMIT,
            mark_read=True,
            offset=0,
            notification_types=None,
            sort_order=SortOrder.ascending,
            start_time=None):
        """
        Return a list of notifications that have been sent to the specified
        recipient.


        :param app_id: Identification of the client application such as a Web,
            a desktop, or a mobile application, that accesses the service.

        :param recipient_id: Identification of a recipient that may have been
            issued notifications, such as, for instance the identification of
            a user account, or the identification of a device, such as an
            International Mobile Equipment Identity (IMEI) for Android device,
            or an alphanumeric string for iOS device.

        :param connection: An object `RdbmsConnection`.  Auto commit MUST be
            enabled if the argument `mark_read` is equal to `True`.

        :param end_time: Indicate the latest time of submission to return
            notifications.

        :param include_read: Indicate whether to include notifications that
            have been already read.

        :param limit: Constrain the number of notifications that are returned
            to the specified number.  Default value is
            `NotificationService.DEFAULT_LIMIT`.  Maximum value is
            is `NotificationService.MAXIMUM_LIMIT`.

        :param mark_read: Indicate whether to mark as read every notification
            that are returned.

        :param notification_types: A list of notification types the recipient
            is interested in, or whatever notification if not defined.

        :param offset: Require to skip that many records before beginning to
            return notifications.  Default value is `0`.  If both `offset`
            and `limit` are specified, then `offset` records are skipped
            before starting to count the limit notifications that are
            returned.

        :param sort_order: Ascending order sorts notifications by ascending
            schedule time, while descending order sorts notifications by
            descending schedule time.

        :param start_time: Indicate the earliest time of submission to return
            notifications.


        :return: A list of instances containing the following members:

            * `creation_time` (required): Time when the sender originated the
              notification to the intended recipient.

            * `is_read` (required): Indicate whether the notification has been read
              by the intended recipient.

            * `notification_id` (required): Identification of the  notification.

            * `notification_type` (required): String representation of the type of
              the notification, as selected by the sender that originated this
              notification to the intended recipient.

            * `payload` (optional): An arbitrary JSON expression added by the
              sender to provide information about the context of this notification.

            * `schedule_time` (required): Time when this notification is scheduled
              to be sent to the intended recipient.  The notification is not visible
              to the intended recipient prior to this time.

            * `sender_id` (optional): The identification of the sender that
              originated the notification.

            * `update_time` (required): Time of the most recent modification of an
              attribute of the notification, such as its read status.
        """
        with self.acquire_rdbms_connection(auto_commit=mark_read, connection=connection) as connection:
            if mark_read:
                cursor = connection.execute(
                    f"""
                    UPDATE 
                        notification
                      SET
                        is_read = true
                      WHERE
                        notification_id IN (
                          SELECT 
                              notification_id
                            FROM
                              notification
                            WHERE
                              recipient_id = %(recipient_id)s
                              AND (%(start_time)s IS NULL OR creation_time > %(start_time)s)
                              AND (%(end_time)s IS NULL OR creation_time < %(end_time)s)
                              AND (%(include_read)s OR NOT is_read)
                              AND (%(notification_types)s IS NULL OR notification_type IN (%(notification_types)s))
                              AND object_status = %(OBJECT_STATUS_ENABLED)s
                            ORDER BY
                              creation_time {'ASC' if sort_order == SortOrder.ascending else 'DESC'}
                            OFFSET %(offset)s
                            LIMIT %(limit)s)
                      RETURNING 
                        notification_id,
                        notification_type,
                        is_read,
                        sender_id,
                        payload,
                        schedule_time,
                        creation_time,
                        update_time
                    """,
                    {
                        'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
                        'end_time': end_time,
                        'include_read': include_read,
                        'limit': min(limit, self.MAXIMUM_LIMIT) or self.DEFAULT_LIMIT,
                        'notification_types': notification_types,
                        'offset': offset,
                        'recipient_id': recipient_id,
                        'start_time': start_time
                    })

            else:
                cursor = connection.execute(
                    f"""
                    SELECT 
                        notification_id,
                        notification_type,
                        is_read,
                        sender_id,
                        payload,
                        schedule_time,
                        creation_time,
                        update_time
                      FROM
                        notification
                      WHERE
                        recipient_id = %(recipient_id)s
                        AND (%(start_time)s IS NULL OR creation_time > %(start_time)s)
                        AND (%(end_time)s IS NULL OR creation_time < %(end_time)s)
                        AND (%(include_read)s OR NOT is_read)
                        AND (%(notification_types)s IS NULL OR notification_type IN (%(notification_types)s))
                        AND object_status = %(OBJECT_STATUS_ENABLED)s
                      ORDER BY
                        creation_time {'ASC' if sort_order == SortOrder.ascending else 'DESC'}
                      OFFSET %(offset)s
                      LIMIT %(limit)s
                    """,
                    {
                        'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
                        'end_time': end_time,
                        'include_read': include_read,
                        'limit': min(limit, self.MAXIMUM_LIMIT) or self.DEFAULT_LIMIT,
                        'notification_types': notification_types,
                        'offset': offset,
                        'recipient_id': recipient_id,
                        'start_time': start_time
                    })

            notifications = [
                row.get_object({
                    'creation_time': cast.string_to_timestamp,
                    'notification_id': cast.string_to_uuid,
                    'payload': cast.string_to_json,
                    'schedule_time': cast.string_to_timestamp,
                    'update_time': cast.string_to_timestamp
                })
                for row in cursor.fetch_all()
            ]

            return notifications

    def register_device(
            self,
            app_id,
            device_id,
            device_token,
            device_platform,
            account_id=None,
            connection=None,
            locale=None,
            topics=None,
            utc_offset=None):
        """
        Register a device to receive push notification messages from the
        platform.


        :note: the function registers the device on behalf of the application
            identification of the server platform, not the identification of
            the client application itself.  An instance of a server platform
            is specific to a particular service to which client applications,
            whatever their platform (Android, iOS, Web, etc.) and therefore
            whatever their application identification, are interested in
            receiving push notifications.


        @todo: the argument "topics" is not used at the moment.


        :param app_id: Identification of the client application such as a Web,
            a desktop, or a mobile application, that accesses the service.

        :param device_id: Identification of the device, which depends on the
            device platform:

            * Android: International Mobile Equipment Identity (IMEI) number
              of the device.

            * iOS: unique identifier of the iOS device, previously the
              Unique Device Identifier (UDID) of the device, which is a
              40-character string that is tied to this specific Apple device.
              It could be a SecureUDID, which is an open-source sandboxed UDID
              solution aimed at solving the main privacy issues that caused
              Apple to deprecate UDIDs.

        :param device_token: Token that identifies the device by the push
            notification provider of the device platform.

            * Android: token identifying the device to push the notification
              to, i.e., the registration ID.  A device token is an opaque
              identifier of a device that Android Google Cloud Messaging (GCM)
              gives to the device when it first connects with it.  The device
              shares the device token with its provider. The device token is
              analogous to a phone number; it contains information that enables
              GCM to locate the device on which the client application is
              installed.  GCM also uses it to authenticate the routing of a
              notification.

            * iOS: token identifying the iOS device to push the notification
              to.  A device token is an opaque identifier of a device that APNs
              gives to the device when it first connects with it.  The device
              shares the device token with its provider. Thereafter, this token
              accompanies each notification from the provider.  The device
              token is analogous to a phone number; it contains information
              that enables APNs to locate the device on which the client
              application is installed. APNs also uses it to authenticate the
              routing of a notification.  A device token is not the same thing
              as the device UDID returned by the `uniqueIdentifier` property
              of `UIDevice`.

        :param device_platform: Indicate the platform of the end user's mobile
            device:

            * `ios`: Apple iOS

            * `android`: Google Android

            * `windows`: Windows Phone

        :param account_id: Identification of the account of the user on behalf
            of whom the device is registered to receive push notification
            messages.

        :param connection: An object `RdbmsConnection` with auto commit.

        :param locale: Represent the language that the end user prefers
            receiving new content in.  A locale corresponds to a tag
            respecting RFC 4646, expressed by a ISO 639-3 alpha-3 code
            element, optionally followed by a dash character `-` and a ISO
            3166-1 alpha-2 code.  For example: "eng" (which denotes a standard
            English), "eng-US" (which denotes an American English).  If this
            argument is not specified, the locale corresponds to the preferred
            language that the user has defined in his profile.

        :param topics: A list of keywords representing topics the end user is
            interested in to be pushed new content whenever related to one of
            those topics.  The list of supported keywords is specific to the
            publisher service of the client application and as such the
            developer of the client application has to refer to the technical
            documentation of the publisher service.

        :param utc_offset: Difference between the location of the device and
            UTC (Universal Time Coordinated).  UTC is also known as GMT or
            Greenwich Mean Time or Zulu Time.


        :return: An object containing the following attributes:

            * `device_id` (required): Identification of the mobile device that
              registers to push notification service.

            * `object_status` (required): Current status of the registration of
              this device to push notification service.

            * `registration_id` (required): Identification of the registration of
              the mobile device to the push notification service.

            * `update_time` (required): Time of the most recent modification of
              information of the mobile device's registration to the push
              notification service.
        """

        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            ApplicationService().get_application(app_id, check_status=True, connection=connection)

            device_registration = self.__get_device_registration(
                device_id,
                app_id,
                account_id=account_id,
                connection=connection)

            if device_registration is None:
                device_registration = self.__add_device_registration(
                    device_id,
                    device_platform,
                    app_id,
                    device_token,
                    account_id=account_id,
                    connection=connection,
                    locale=locale,
                    utc_offset=utc_offset)
            elif device_registration.device_token != device_token:
                device_registration = self.__update_device_registration_token(
                    device_registration.registration_id,
                    device_token,
                    connection=connection)

            return device_registration

    def send_notification(
            self,
            app_id,
            notification_type,
            connection=None,
            recipient_ids=None,
            team_id=None,
            if_not_exists=False,
            lifespan=DEFAULT_LIFESPAN,
            notification_mode=None,
            package=None,
            payload=None,
            sender_id=None,
            schedule_time=None,
            recipient_timezones=None):
        """
        Send a notification to the intended recipient(s) as soon as possible
        or at a given time.

        The notification and its content is stored for later delivery up to a
        maximum period of time, the lifespan of the notification, also known
        as its time-to-live(TTL).  The primary reason for this is that a
        device may be unavailable (e.g., turned off, no network coverage).


        :param app_id: identification of the client application such as a Web,
            a desktop, or a mobile application, that accesses the service.

        :param notification_type: type of the notification to be sent, such as
            for instance `on_something_happened`.

        :param connection: a `RdbmsConnection` instance to be used
            supporting the Python clause `with ...:`.

        :param recipient_ids: identification of a recipient, or a list of
            identifications of recipients, to send the notification to.

        :param team_id: identification of a team to send the notification to
            all the members.

        :param if_not_exists: indicate whether the notification needs to be
            ignored for a particular recipient if the latter already received
            a notification of the same type that the recipient has not yet
            read.  The function doesn't check whether a previous notification
            of the same type may have a different payload.

        :param lifespan: period of time expressed in seconds during which
            information of the notification reasonably may be expected
            usefully to inform an action or interest of the intended
            recipients.  After the lifespan expires, provision of the
            notification to the recipients may be prevented; the notification
            and its content may be deleted.  The maximum lifespan is
            `NotificationService.MAXIMUM_LIFESPAN`.

        :param notification_mode: an item of the enumeration `NotificationMode`
            to indicate the mode to deliver this notification to the
            recipients:

            * `pull` (default): indicate that a notification message is
              delivered to the specified recipients when the request for the
              transmission of information is initiated by the receiver or client
              application, and then is responded by the publisher or server
              platform.

            * `push`: indicate that a notification message is delivered to the
              specified recipients when the request for the transmission of
              information is initiated by the publisher or server platform and
              pushed out to the receiver or client application.

        :param payload: any arbitrary JSON expression added by the sender to
            provide information about the context of this notification.

        :param sender_id: identification of the sender on behalf whom the
            notification is sent to the recipients.

        :param schedule_time: Time when this notification needs to be sent to
            the intended recipient.  The notification is not visible to the
            intended recipient prior to this time.  If not specified, the
            notification is sent as soon as possible.

        :param recipient_timezones: dictionary of time zones of the intended
            recipients for which the schedule time is expected to be in their
            local time.  The key corresponds to the identification of an
            intended recipient, and the value corresponds to time zone of this
            particular recipient.

            When a time zone is defined for a particular recipient, the
            function understands that the schedule time of the notification
            for this recipient has to be converted to the local time of this
            recipient.  The function converts `schedule_time` to UTC, and it
            strips the time zone information to get a local time to which the
            function adds the recipient's time zone.

            For instance, if the specified schedule time is
            `2013-12-20 15:00:00+07` and the time zone `+7` is specified
            for a recipient, the schedule time is assumed to be the local time
            `2013-12-20 08:00:00` for the region, i.e., time zone, of this
            recipient.  The resulting schedule time with time zone for this
            recipient is then `2013-12-20 08:00:00+07`


        :return: a list of instances containing the following members:

            * `notification_id`: Identification of the notification sent to a
              particular recipient.

            * `recipient_id`: String representation of the identification of
              the recipient intended to be delivered this notification.
        """
        if not recipient_ids and team_id is None:
            return

        # Convert the argument `recipient_ids` to a list if it's not already
        # a list, a set, or a tuple.
        recipient_ids = [] if recipient_ids is None \
            else set([str(recipient_ids)] if not isinstance(recipient_ids, (list, set, tuple)) \
            else [str(recipient_id) for recipient_id in recipient_ids])

        # If an organisation has been specified, retrieve the list of all its
        # members to send the notification to.
        if team_id:
            offset = 0

            while True:
                accounts = TeamService().get_members(
                    app_id,
                    team_id,
                    limit=TeamService.DEFAULT_LIMIT,
                    offset=offset)

                recipient_ids.update([str(account.account_id) for account in accounts])
                if len(accounts) < TeamService.DEFAULT_LIMIT:
                    break

                offset += len(accounts)

        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            # Filter out the recipients that have already been sent this
            # notification type but they have not read it yet.
            if if_not_exists:
                cursor = connection.execute(
                    """
                    SELECT DISTINCT 
                        recipient_id
                      FROM 
                        notification
                      WHERE
                        notification_type = %(notification_type)s
                        AND recipient_id IN (%[recipient_ids]s)
                        AND (%(sender_id)s IS NULL OR sender_id = %(sender_id)s)
                        AND NOT is_read
                    """,
                    {
                        'notification_type': notification_type,
                        'recipient_ids': recipient_ids,
                        'sender_id': sender_id
                    })

                recipient_ids = set(recipient_ids) - set([ row.get_value('recipient_id') for row in cursor.fetch_all() ])

                if len(recipient_ids) == 0:
                    return

            expiration_time = ISO8601DateTime.now() + datetime.timedelta(
                seconds=max(lifespan or self.DEFAULT_LIFESPAN, self.MAXIMUM_LIFESPAN))

            print('######################')
            print(type(obj.stringify(payload)))
            print(str(payload))
            print(type(str(payload)))
            print('######################')

            # Register this notification for every recipient.
            cursor = connection.execute(
                """
                INSERT INTO notification(
                    notification_type,
                    notification_mode,
                    recipient_id,
                    sender_id,
                    payload,
                    schedule_time,
                    expiration_time,
                    app_id)
                  VALUES
                    %[values]s
                  RETURNING
                    notification_id,
                    recipient_id
                """,
                {
                    'values': [
                        (
                            notification_type,
                            notification_mode or NotificationService.NotificationMode.push,
                            recipient_id,
                            sender_id,
                            payload and str(payload),
                            schedule_time,
                            expiration_time,
                            # ('DEFAULT', ) if schedule_time is None
                            #     else schedule_time if (recipient_timezones is None or recipient_timezones.get(recipient_id) is None)
                            #     else cast.string_to_timestamp('%s%+03d' % (str(schedule_time)[:-6], recipient_timezones[recipient_id])),
                            # (("current_timestamp + '%d seconds'::interval" % max(lifespan, self.MAXIMUM_LIFESPAN)), ),
                            app_id,
                        )
                        for recipient_id in recipient_ids
                    ]
                })

            return [
                row.get_object({
                    'notification_id': cast.string_to_uuid
                })
                for row in cursor.fetch_all()
            ]

    def unregister_device(
            self,
            app_id,
            device_id,
            account_id=None):
        """
        Unregister a device from receiving push notification messages from the
        platform.


        :param app_id: identification of the client application such as a Web,
            a desktop, or a mobile application, that accesses the service.

        :param device_id: identification of the device, which depends on the
            device platform:

            * Android: International Mobile Equipment Identity (IMEI) number
              of the device.

            * iOS: unique identifier of the iOS device, previously the
              Unique Device Identifier (UDID) of the device, which is a
              40-character string that is tied to this specific Apple device.
              It could be a SecureUDID, which is an open-source sandboxed UDID
              solution aimed at solving the main privacy issues that caused
              Apple to deprecate UDIDs.

        :param account_id: identification of the account of the user on behalf
            of whom the device is unregistered from receiving push
            notification messages.
        """
        with self.acquire_rdbms_connection(True) as connection:
            connection.execute("""
                DELETE FROM 
                    notification_device
                  WHERE 
                    device_id = %(device_id)s
                    AND ((account_id IS NULL AND %(account_id)s IS NULL) OR
                         (account_id = %(account_id)s))
                    AND app_id = %(app_id)s
                """,
                {
                    'account_id': account_id,
                    'app_id': app_id,
                    'device_id': device_id
                })








# def flush_notifications(self, app_id, account_id,
#         service_name=None, notification_types=None):
#     """
#     Flush all the notifications originated from the given client
#     application that were sent to the specified user.
#
#     :param app_id: identification of the client application such as a Web,
#         a desktop, or a mobile application, that accesses the service.
#     :param account_id: identification of the account of the user to flush
#         all the notifications he receives from the given application.
#     :param service_name: code name of the service that originated the
#         notifications to flush.  By convention, the code name of the
#         service corresponds to the Python class of this service (cf.
#         `self.__class__.__name__`).
#     :param notification_types: a list of types of the notifications to
#         flush.  If no list is provided, the function flushes all the
#         notifications that the client application posted to the user.
#     """
#     with self.acquire_rdbms_connection(True) as connection:
#         if notification_types is None:
#             connection.execute("""
#                 DELETE FROM notification
#                   WHERE recipient_id = %(account_id)s
#                     AND app_id = %(app_id)s
#                     AND object_status = %(OBJECT_STATUS_ENABLED)s
#                     AND (%(service_name) IS NULL OR service_name = %(service_name)s)""",
#                 { 'OBJECT_STATUS_DISABLED': OBJECT_STATUS_DISABLED,
#                   'OBJECT_STATUS_ENABLED': OBJECT_STATUS_ENABLED,
#                   'app_id': app_id,
#                   'recipient_id': account_id,
#                   'service_name': service_name })
#         else:
#             connection.execute("""
#                 DELETE FROM notification
#                   WHERE recipient_id = %(account_id)s
#                     AND app_id = %(app_id)s
#                     AND object_status = %(OBJECT_STATUS_ENABLED)s
#                     AND (%(service_name) IS NULL OR service_name = %(service_name)s)
#                     AND notification_type IN (%[notification_types]s)""",
#                 { 'OBJECT_STATUS_DISABLED': OBJECT_STATUS_DISABLED,
#                   'OBJECT_STATUS_ENABLED': OBJECT_STATUS_ENABLED,
#                   'app_id': app_id,
#                   'notification_types': notification_types,
#                   'recipient_id': account_id,
#                   'service_name': service_name })

# def mark_notifications(self, app_id, account_id, notification_ids,
#         mark_read=True):
#     """
#     Update the read mark of a list of notifications sent to the specified
#     user.
#
#     :param app_id: identification of the client application such as a Web,
#            a desktop, or a mobile application, that accesses the service.
#
#     :param account_id: identification of the account of a user whom the
#            specified notifications have been sent to.
#
#     :param notification_ids: a list of notification to mark as read or
#            unread.
#
#     :param mark_read: indicate whether to mark as read the specified
#         notifications.
#     """
#     if notification_ids is None:
#         return
#
#     if type(notification_ids) not in (list, set, tuple):
#         notification_ids = [ notification_ids ]
#
#     if len(notification_ids) == 0:
#         return
#
#     with self.acquire_rdbms_connection(True) as connection:
#         cursor = connection.execute("""
#             UPDATE notification
#               SET is_read = %(mark_read)s,
#                   update_time = current_timestamp
#               WHERE notification_id IN (%[notification_ids]s)
#                 AND recipient_id = %(account_id)s
#                 AND is_read <> %(mark_read)s
#               RETURNING notification_id""",
#             { 'account_id': account_id,
#               'mark_read': mark_read,
#               'notification_ids': notification_ids })
#         missing_notification_ids = set(notification_ids) \
#             - set([ row.get_value('notification_id', cast.string_to_uuid) for row in cursor.fetch_all() ])
#         if len(notification_ids) > 0:
#             raise self.UndefinedObjectException('Some specified notifications have not been sent to this user or their state do not change',
#                 payload=missing_notification_ids)


#
#     def _post_notification(self, app_id, notification_type,
#             notification_mode=NotificationMode.pull,
#             account_ids=None, device_ids=None,
#             payload=None, lifespan=None,
#             topics=None,
#             schedule_time=None, use_local_time=False,
#             is_broadcast=False, is_unique=False, is_volatile=False, is_product_based=False):
#         """
#
#
# The function checks whether the  payload, if any provided, is of a
# simple type such as a number (integer, decimal, complex) or a string.
# If not, the function convert the payload to a string representation
# of a JSON expression.
#
# @warning: this function is for internal usage only; it MUST not be
#     surfaced to any client applications through a RESTful API.
#
# :param app_id: identification of the client application such as a Web,
#     a desktop, or a mobile application, that accesses the service.
# :param notification_type: type of the notification such as
#     `on_something_happened`.
# :param notification_mode:
#
# :param account_ids: list of account identifications of users to send
#     the notification to.
# :param device_ids: list of identifications of devices to send the
#     notification to.
#
# :param payload: content of the notification to send to the recipients.
#     The content could be a simple string caption, or an object which
#     the function converts its JSON expression.into a string
#     representation.
# :param lifespan: duration in minutes the notification lives before it
#     is deleted.  If not defined, the notification persists forever as
#     long as it is not read.
# :param topics: a list of keywords indicating the subjects of this
#     notifications.  Only the subscribers who have registered for these
#     topics will be pushed this notification.
#
# :param schedule_time: schedule the notification to automatically be
#     sent later at the given time.  If not specified, the notification
#     is sent as soon as possible.
# :param use_local_time: indicate whether the schedule time is assumed
#     to be in local time of a particular device.  If so, the
#     `schedule_time` is converted to UTC, and the time zone
#     information is then stripped out to provide a local time.  For
#     instance, if the specified schedule time is `2013-12-20 13:00:00+07`
#     and the argument `use_local_time` is set to `True`, the
#     schedule time is assumed to be the local time `2013-12-20 06:00:00`
#     for the region, i.e., time zone, of a particular device to send
#     the notification to.
#
# :param is_broadcast: indicate whether the notification needs to be
#     sent to every users who subscribed for receiving notification from
#     this application (determined by `app_id`), or more generally
#     from the product this application belongs to (if the argument
#     `is_product_based` is `True`).
#         """
#
#         if account_ids is None and device_ids is None:
#             raise self.InvalidArgumentException('No recipient has been specified')
#
#         if schedule_time and expiration_time and schedule_time > expiration_time:
#             raise self.InvalidArgumentException('The expiration time of a message MUST be posterior to the specified schedule time')
#
#         # If the payload of the message is not a simple type, convert it to a
#         # string representation of a JSON expression.
#         #
#         # :note: `basestring` includes `str` and `unicode`.
#         if payload and not isinstance(payload,  (basestring, int, long, float, complex)):
#             payload = obj.jsonify(payload)


# def send_device_notification(self, app_id, notification_type, device_ids,
#         notification_mode=NotificationMode.push,
#         if_not_exists=False,
#         lifespan=DEFAULT_LIFESPAN,
#         payload=None,
#         sender_id=None):
#     """
#     Send a notification to the intended device(s) as soon as possible or
#     at a given time.
#
#     The notification and its content is stored for later delivery up to a
#     maximum period of time, the lifespan of the notification, also known
#     as its time-to-live(TTL).  The primary reason for this is that a
#     device may be unavailable (e.g., turned off, no network coverage).
#
#
#     :param app_id: identification of the client application such as a Web,
#         a desktop, or a mobile application, that accesses the service.
#
#     :param notification_type: type of the notification to be sent, such as
#         for instance `on_something_happened`.
#
#     :param device_ids: identification of the device, or a list of
#         identifications of devices, to send the notification to.
#
#     :param if_not_exists: indicate whether the notification needs to be
#         ignored for a particular device if the latter already received a
#         notification of the same type that the device has not yet read.
#         The function doesn't check whether a previous notification of the
#         same type may have a different payload.
#
#     :param lifespan: period of time expressed in seconds during which
#         information of the notification reasonably may be expected
#         usefully to inform an action or interest of the intended devices.
#         After the lifespan expires, provision of the notification to the
#         devices may be prevented; the notification and its content may be
#         deleted.  The maximum lifespan is `NotificationService.MAXIMUM_LIFESPAN`.
#
#     :param payload: any arbitrary JSON expression added by the sender to
#         provide information about the context of this notification.
#
#     :param sender_id: identification of the sender on behalf whom the
#         notification is sent to the devices.
#
#
#     :return: a list of instances containing the following members:
#
#         * `creation_time` (required): time when the notification was
#           registered to the platform.
#
#         * `device_id` (required): identification of a device that is sent
#           this notification.
#
#         * `notification_id` (required): identification of the notification
#           sent to the device.
#     """
#     if device_ids is None:
#         return
#
#     if not isinstance(device_ids, (list, set, tuple)):
#         device_ids = [ device_ids ]
#
#     with self.acquire_rdbms_connection(True) as connection:
#         # Filter out from the list of the devices to send this notification
#         # those that have received this notification type but not read yet.
#         if if_not_exists == True:
#             cursor = connection.execute("""
#                 SELECT DISTINCT device_id
#                   FROM notification2device
#                   WHERE device_ud IN (%[device_ids]s)
#                     AND (%(account_id)s IS NULL OR account_id = %(account_id)s)
#                     AND notification_type = %(notification_type)s
#                     AND is_unread
#                     AND (expiration_time IS NULL OR expiration_time > current_timestamp)""",
#                 { 'notification_type': notification_type,
#                   'device_ids': device_ids,
#                   'account_id': sender_id })
#             device_ids = set(device_ids) - set([ row.get_value('device_id') for row in cursor.fetch_all() ])
#             if len(device_ids) == 0:
#                 return
#
#         cursor = connection.execute("""
#             INSERT INTO notification2device(
#                   notification_type,
#                   notification_mode,
#                   device_id,
#                   sender_id,
#                   payload,
#                   expiration_time,
#                   app_id)
#               VALUES %[values]s
#               RETURNING notification_id,
#                         device_id,
#                         creation_time""",
#             { 'values': [ (notification_type,
#                            notification_mode,
#                            device_id,
#                            sender_id,
#                            payload and obj.jsonify(payload),
#                            lifespan and (True, "current_timestamp + '%d seconds'::interval" % lifespan),
#                            app_id)
#                     for device_id in device_ids ] })
#
#         return [ row.get_object({
#                         'creation_time': cast.string_to_timestamp,
#                         'notification_id': cast.string_to_uuid })
#                 for row in cursor.fetch_all() ]


# def send_notifications(self, app_id, notification_type, recipient_ids, payloads,
#         if_not_exists=False,
#         lifespan=DEFAULT_LIFESPAN,
#         notification_mode=NotificationMode.pull,
#         package=None,
#         sender_id=None,
#         schedule_time=None,
#         recipient_timezones=None):
#
#     if not recipient_ids:
#         return
#
#     recipient_ids = [ str(recipient_ids) ] if not isinstance(recipient_ids, (list, set, tuple)) \
#             else [ str(recipient_id) for recipient_id in recipient_ids ]
#
#     json_payloads = [ obj.jsonify(payload) for payload in payloads ]
#
#     with self.acquire_rdbms_connection(True) as connection:
#         if if_not_exists:
#             cursor = connection.execute("""
#                 SELECT DISTINCT recipient_id
#                   FROM notification
#                   WHERE notification_type = %(notification_type)s
#                     AND recipient_id IN (%[recipient_ids]s)
#                     AND (%(sender_id)s IS NULL OR sender_id = %(sender_id)s)
#                     AND NOT is_read""",
#                 { 'notification_type': notification_type,
#                   'recipient_ids': recipient_ids,
#                   'sender_id': sender_id })
#
#             recipient_ids = set(recipient_ids) - set([ row.get_value('recipient_id') for row in cursor.fetch_all() ])
#
#             if len(recipient_ids) == 0:
#                 return
#
#         cursor = connection.execute("""
#             INSERT INTO notification(
#                 notification_type,
#                 notification_mode,
#                 recipient_id,
#                 sender_id,
#                 payload,
#                 schedule_time,
#                 expiration_time,
#                 app_id,
#                 package)
#               VALUES %[values]s
#               RETURNING notification_id,
#                         recipient_id""",
#             { 'values': [
#                     (notification_type,
#                      notification_mode,
#                      recipient_id,
#                      sender_id,
#                      json_payload,
#                      (True, 'DEFAULT') if schedule_time is None \
#                         else schedule_time if (recipient_timezones is None or recipient_timezones.get(recipient_id) is None) \
#                         else cast.string_to_timestamp('%s%+03d' % (str(schedule_time)[:-6] ), recipient_timezones[recipient_id]),
#                      (True, "current_timestamp + '%d seconds'::interval" % max(lifespan, self.MAXIMUM_LIFESPAN)),
#                      app_id,
#                      package)
#                     for json_payload in json_payloads
#                         for recipient_id in recipient_ids ] })
#
#         return [ row.get_object({ 'notification_id': cast.string_to_uuid })
#                 for row in cursor.fetch_all() ]




# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging

import settings


class BaseService:
    """
    Base class of service.

    It provides logging facility to the inheriting service class.  The
    logger is identifier by the global constant `LOGGER_NAME` of
    the Python module `settings`.

    It also provides facility to publish messages to an Advanced
    Message Queuing Protocol (AMQP) compliant broker, which connection
    properties are defined in the dictionary
    `AMQP_CONNECTION_PROPERTIES` defined in the Python module.
    `settings`.
    """
    class BaseServiceException(Exception):
        """
        Base class of exception that a service might raise to indicate some
        unexpected conditions.
        """
        def __init__(self, message=None, payload=None):
            """
            Build a new instance of an exception that indicates an unexpected
            condition met by a service when processing a client's request.


            :param message: A humanly-readable message that describes this
                exception instance.

            :param payload: A JSON expression that provides detailed information
                about this exception instance.
            """
            super().__init__(message)
            self.payload = payload

    class DeletedObjectException(BaseServiceException):
        """
        Signal that the specified object has been deleted and as such it
        cannot be used or referenced anymore.
        """

    class DisabledObjectException(BaseServiceException):
        """
        Signal that the specified object has been disabled and as such it
        cannot be used.
        """

    class InvalidOperationException(BaseServiceException):
        """
        Signal the operation requested is invalid, for instance one or several
        arguments provided within the request are not coherent together or the
        state of the object on which the operation is requested.
        """

    class IllegalAccessException(BaseServiceException):
        """
        Signal that the user on behalf whom a function is called is not
        authorized to perform the action requested.
        """

    class InvalidArgumentException(BaseServiceException):
        """
        Signal that an argument passed to a function is invalid, more likely
        of the wrong type.
        """

    class PendingObjectException(BaseServiceException):
        """
        Signal that the specified object is still pending and as such it
        cannot be used until it has been enabled.
        """

    class UndefinedObjectException(BaseServiceException):
        """
        Signal that the specified object doesn't exist and as such it cannot
        be referenced.
        """

    # Logger for custom service trace.
    __logger__ = logging.getLogger(getattr(settings, 'LOGGER_NAME'))

    def __init__(self):
        self._service_name = self.__class__.__name__

    def get_service_name(self):
        return self._service_name

    @staticmethod
    def log_critical(message):
        BaseService.__logger__.critical(message)

    @staticmethod
    def log_debug(message):
        BaseService.__logger__.debug(message)

    @staticmethod
    def log_error(message):
        BaseService.__logger__.error(message)

    @staticmethod
    def log_info(message):
        BaseService.__logger__.info(message)

    @staticmethod
    def log_warning(message):
        BaseService.__logger__.warning(message)



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.constant.http import HttpMethod
from majormode.perseus.service.base_http_handler import HttpRequest
from majormode.perseus.service.base_http_handler import HttpRequestHandler
from majormode.perseus.service.base_http_handler import http_request
from majormode.perseus.service.team.team_service import TeamService


class TeamServiceHttpRequestHandler(HttpRequestHandler):
    @http_request(r'^/team/invitation/(invitation_code)$',
                  http_method=HttpMethod.PUT,
                  authentication_required=False)
    def accept_or_decline_invitation(self, request, invitation_code):
        does_accept = request.get_argument('accept',
                argument_passing=HttpRequest.ArgumentPassing.query_string,
                data_type=HttpRequest.ArgumentDataType.boolean,
                is_required=True)
        return TeamService().accept_invitation(request.app_id, invitation_code, account_id=request.account_id) if does_accept \
                else TeamService().decline_invitation(request.app_id, invitation_code, account_id=request.account_id)

    @http_request(r'^/team/invitation/(invitation_id:uuid)$',
                  http_method=HttpMethod.DELETE,
                  authentication_required=True)
    def cancel_invitation(self, request, invitation_id):
        return TeamService().cancel_invitation(request.app_id, request.account_id, invitation_id)

    @http_request(r'^/team$',
                  http_method=HttpMethod.POST,
                  authentication_required=True)
    def create_team(self, request):
        name = request.get_argument('name',
                data_type=HttpRequest.ArgumentDataType.string,
                is_required=True)
        description = request.get_argument('description',
                data_type=HttpRequest.ArgumentDataType.string,
                is_required=False)
        invitation_url = request.get_argument('invitation_url',
                data_type=HttpRequest.ArgumentDataType.string,
                is_required=False)
        invitation_email = request.get_argument('invitation_email',
                data_type=HttpRequest.ArgumentDataType.string,
                is_required=False)
        return TeamService().create_team(request.app_id, request.account_id, name,
                description=description,
                invitation_url=invitation_url,
                invitation_email=invitation_email)

    @http_request(r'^/team/(team_id:uuid)/invitation$',
                  http_method=HttpMethod.GET,
                  authentication_required=True)
    def get_invitations(self, request, team_id):
        offset = request.get_argument('offset',
                data_type=HttpRequest.ArgumentDataType.integer,
                is_required=False,
                default_value=0)
        limit = request.get_argument('limit',
                data_type=HttpRequest.ArgumentDataType.integer,
                is_required=False,
                default_value=TeamService.DEFAULT_LIMIT)
        return TeamService().get_invitations(request.app_id, request.account_id, team_id,
                offset=offset, limit=limit)

    @http_request(r'^/team/(team_id:uuid)/member/(account_id:uuid)$',
                  http_method=HttpMethod.GET,
                  authentication_required=True)
    def get_member(self, request, team_id, account_id):
        return TeamService().get_member(request.app_id, request.account_id, team_id, account_id)

    # @http_request(r'^/team/(team_id:uuid)/member$',
    #               http_method=HttpMethod.GET,
    #               authentication_required=True)
    # def get_members(self, request, team_id):
    #     offset = request.get_argument('offset',
    #             data_type=HttpRequest.ArgumentDataType.integer,
    #             is_required=False,
    #             default_value=0)
    #     limit = request.get_argument('limit',
    #             data_type=HttpRequest.ArgumentDataType.integer,
    #             is_required=False,
    #             default_value=TeamService.DEFAULT_LIMIT)
    #     return TeamService().get_members(request.app_id, request.account_id, team_id,
    #             offset=offset, limit=limit)

    @http_request(r'^/team$',
                  http_method=HttpMethod.GET,
                  authentication_required=False)
    def get_teams(self, request):
        keywords = request.get_argument('keywords',
                data_type=HttpRequest.ArgumentDataType.list,
                item_data_type=HttpRequest.ArgumentDataType.string,
                is_required=False)

        limit = request.get_argument('limit',
                data_type=HttpRequest.ArgumentDataType.integer,
                is_required=False,
                default_value=TeamService.DEFAULT_LIMIT)

        offset = request.get_argument('offset',
                data_type=HttpRequest.ArgumentDataType.integer,
                is_required=False,
                default_value=0)

        return TeamService().search_teams(request.app_id, request.account_id, keywords,
                limit=limit, offset=offset) if keywords \
            else TeamService().get_account_teams(request.app_id, request.account_id,
                limit=limit, offset=offset)

    @http_request(r'^/team/(team_id:uuid)/invitation$',
                  http_method=HttpMethod.POST,
                  authentication_required=True)
    def invite_users(self, request, team_id):
        if not isinstance(request.body, list):
            raise TeamService.InvalidArgumentException('Require a list of users identified either by their account identifications, either their email address')
        return TeamService().invite_users(request.app_id, request.account_id, team_id, request.body)

    @http_request(r'^/team/(team:uuid)/agent/(account_id:uuid)$',
                  http_method=HttpMethod.PUT,
                  authentication_required=True)
    def promote_agent(self, request, team_id, account_id):
        return TeamService().promote_agent(request.app_id, request.account_id, team_id, account_id)

    @http_request(r'^/team/(team_id:uuid)/member/(account_id:uuid)$',
                  http_method=HttpMethod.DELETE,
                  authentication_required=True)
    def revoke_member(self, request, team_id, account_id):
        return TeamService().revoke_member(request.app_id, request.account_id, team_id, account_id)

    @http_request(r'^/team/(team_id:uuid)/join$',
                  http_method=HttpMethod.POST,
                  authentication_required=True)
    def submit_join_request(self, request, team_id):
        return TeamService().request_membership(request.app_id, request.account_id, team_id)

    @http_request(r'^/team/(team_id:uuid)/member/(account_id:uuid)/role$',
                  http_method=HttpMethod.PUT,
                  authentication_required=True)
    def update_member_role(self, request, team_id, account_id):
        is_administrator = request.get_argument('is_administrator',
                data_type=HttpRequest.ArgumentDataType.boolean)
        return TeamService().update_member_role(request.app_id, request.account_id,
                team_id, account_id, is_administrator)

    @http_request(r'^/team/(team_id:uuid)/membership$',
                  http_method=HttpMethod.DELETE,
                  authentication_required=True)
    def withdraw_membership(self, request, team_id):
        return TeamService().withdraw_membership(request.app_id, request.account_id, team_id)



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.service.account.account_service import AccountService
from majormode.perseus.unittest.base_unittest import BaseApiKeyTestCase

import hashlib
import settings
import unittest
import uuid

class AccountServiceTestCase(BaseApiKeyTestCase):
    @staticmethod
    def generate_account(app_id=settings.PLATFORM_UNITTEST_APP_ID,
            account_type=AccountService.AccountType.test):
        nonce = uuid.uuid1().hex
        account = AccountService().sign_up(app_id, 'test_%s@example.com' % nonce,
            fullname='Test %s' % nonce,
            password=nonce[:16],
            account_type=account_type)
        return AccountService().get_account(account.account_id)

    @staticmethod
    def generate_accounts(app_id, size, reused=False):
        if reused == False:
            accounts = [ AccountServiceTestCase.generate_account(app_id) for _ in range(size) ]
        else:
            accounts = []
            for i in range(size):
                email_address = 'test_%s@example.com' % i
                try:
                    account = AccountService().get_account_by_contact_information(
                        (AccountService.VCardPropertyName.EMAIL, email_address))
                except AccountService.UndefinedObjectException:
                    account = AccountService().sign_up(app_id, email_address,
                        fullname='Test %s' % i,
                        password=uuid.uuid1().hex[:16],
                        account_type=AccountService.AccountType.test)
                accounts.append(account)

        return accounts

    @staticmethod
    def generate_credentials():
        nonce = uuid.uuid1().hex
        return ('Test %s' % nonce,
                'test_%s@example.com' % nonce,
                nonce[:16])

    def test_sign_up_with_invalid_credentials(self):
        """
        Test the method ``sign_up`` providing invalid credentials, i.e.,
        several combinations of email address and password that don't respect
        the requirements.
        """
        INVALID_CREDENTIALS = [
            # Email address doesn't respect requirements.
            ('Joe Bar', '%s@examplecom', 'K57p2N6p', 'email address must include top-level domain (TLD)'),
            ('Joe Bar', '%s@example.', 'K57p2N6p', 'email address must include valid top-level domain (TLD)'),
            ('Joe Bar', '%s@.com', 'K57p2N6p', 'email address must include a subdomain name'),

            # Password doesn't respect requirements.
            ('Joe Bar', '%s@example.com', 'bar', 'password must be longer that 6 characters'),
            ('Joe Bar', '%s@example.com', '1234567890123456789', 'password must shorter than 18 characters'),
        ]

        for (fullname, email_address, password, failure_reason) in INVALID_CREDENTIALS:
            email_address = email_address % uuid.uuid1().hex
            try:
                account = AccountService().sign_up(self.unittest_account_id, email_address,
                    fullname=fullname, password=password,
                    account_type=AccountService.AccountType.test)
                self.assertTrue(False, failure_reason)
            except AccountService.InvalidArgumentException:
                pass

    def test_sign_up_with_valid_credentials(self):
        (fullname, email_address, password) = self.generate_credentials()
        account = AccountService().sign_up(self.unittest_account_id, email_address,
            fullname=fullname, password=password,
            account_type=AccountService.AccountType.test)

    def test_sign_up_with_duplicate_email_address(self):
        (fullname, email_address, password) = self.generate_credentials()
        account = AccountService().sign_up(self.unittest_account_id, email_address,
            fullname=fullname, password=password,
            account_type=AccountService.AccountType.test)
        self.assertRaises(AccountService.EmailAddressAlreadyRegisteredException, AccountService().sign_up,
                self.unittest_app_id, email_address,
                fullname=fullname, password=password,
                account_type=AccountService.AccountType.test)

    def test_sign_in_with_invalid_credentials(self):
        (fullname, email_address, password) = self.generate_credentials()
        AccountService().sign_up(self.unittest_app_id, email_address,
            fullname=fullname, password=password,
            account_type=AccountService.AccountType.test)
        self.assertRaises(AccountService.AuthenticationFailureException, AccountService().sign_in,
            self.unittest_app_id, email_address, hashlib.md5(password).hexdigest())

    def test_sign_in_with_valid_credentials(self):
        (fullname, email_address, password) = self.generate_credentials()
        account = AccountService().sign_up(self.unittest_account_id, email_address,
            fullname=fullname, password=password,
            account_type=AccountService.AccountType.test)
        AccountService().sign_in(self.unittest_app_id, email_address, password)

    def test_sign_out(self):
        (fullname, email_address, password) = self.generate_credentials()
        account = AccountService().sign_up(self.unittest_account_id, email_address,
            fullname=fullname, password=password,
            account_type=AccountService.AccountType.test)
        session = AccountService().sign_in(self.unittest_app_id, email_address, password)
        AccountService().sign_out(self.unittest_app_id, account.account_id, session.session_id)

    def test_sign_out_with_invalid_session(self):
        (fullname, email_address, password) = self.generate_credentials()
        account = AccountService().sign_up(self.unittest_account_id, email_address,
            fullname=fullname, password=password,
            account_type=AccountService.AccountType.test)
        AccountService().sign_in(self.unittest_app_id, email_address, password)
        self.assertRaises(AccountService.IllegalAccessException, AccountService().sign_out,
            self.unittest_app_id, account.account_id, uuid.uuid1().hex)

    def test_sign_out_with_invalid_account(self):
        (fullname, email_address, password) = self.generate_credentials()
        account = AccountService().sign_up(self.unittest_account_id, email_address,
            fullname=fullname, password=password,
            account_type=AccountService.AccountType.test)
        session = AccountService().sign_in(self.unittest_app_id, email_address, password)
        self.assertRaises(AccountService.UndefinedObjectException, AccountService().sign_out,
            self.unittest_app_id,uuid.uuid1().hex, session.session_id)

def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.makeSuite(AccountServiceTestCase))
    return test_suite

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# from majormode.perseus.model.obj import OBJECT_STATUS_DELETED
# from majormode.perseus.model.obj import OBJECT_STATUS_ENABLED
# from majormode.perseus.model.obj import OBJECT_STATUS_PENDING
# from majormode.perseus.model.obj import Object
# from majormode.perseus.service.notification.notification_service import NotificationService
#
# from majormode.utils import cast
# from majormode.utils.agent import AbstractAgent
# from majormode.utils.date_util import ISO8601DateTime
# from majormode.utils.rdbms import RdbmsConnection
#
# import argparse
# import blist
# import datetime
# import os
# import settings
# import sys
# import threading
# import time
# import traceback
#
#
# PUSH_NOTIFICATION_BATCH_SIZE = 100
#
# # Define the default number of worker threads that will be used to
# # process new photos that have been captured and uploaded to the local
# # server box.
# DEFAULT_WORKER_POOL_SIZE = 1
#
#
# class PushNotificationController(AbstractAgent):
#     # Semaphore to protect concurrent accesses to the queue of push
#     # notification notifications from worker threads.
#     SEMAPHORE_NOTIFICATIONS = threading.Semaphore()
#
#     NOTIFICATION_SERVICE = NotificationService()
#
#     # Queue of ``PushNotification` instances corresponding to
#     # notifications which need to be sent to a particular device as soon as
#     # possible or in a close future according to its schedule time.
#     # These notifications are sorted in their ascending order of schedule time;
#     # notifications with no schedule time defined come first (as a fortunate
#     # consequence of comparing `None` with an object non-null).
#     notifications = blist.sortedset(key=lambda x: x.schedule_time)
#
#
#     class __PushNotificationAgent__(AbstractAgent):
#         """
#         Class responsible for pushing at a given time one notification to one
#         device, independently from other workers.
#         """
#         def __init__(self):
#             """
#             Build a ``__PushNotificationAgent__`` instance.
#             """
#             super(PushNotificationController.__PushNotificationAgent__, self).__init__()
#             self.notification = None
#
#         def do(self):
#             if self.notification is None:
#                 return False
#
#             try:
#                 _notification_ = Object(
#                         creation_time=self.notification.creation_time,
#                         notification_id=self.notification.notification_id,
#                         notification_type=self.notification.notification_type,
#                         payload=self.notification.payload,
#                         schedule_time=self.notification.schedule_time,
#                         sender_id=self.notification.sender_id)
#
#                 if self.notification.device_platform == 'ios':
#                     PushNotificationController.NOTIFICATION_SERVICE._push_notification_to_ios(
#                             self.notification.device_token,
#                             alert=_notification_)
#
#                 elif self.notification.device_platform == 'android':
#                     PushNotificationController.NOTIFICATION_SERVICE._push_android_notification(
#                             settings.GCM_API_KEY,
#                             self.notification.device_token,
#                             _notification_,
#                             #delay_while_idle=True, # Indicate that the message should not be sent until the device becomes active.
#                             time_to_live=self.notification.time_to_live)
#
#                 sys.stdout.write('#') ; sys.stdout.flush()
#
#                 self.notification = None
#
#             except:
#                 print traceback.format_exc()
#
#             return True
#
#
#         def is_available(self):
#             """
#             Indicate whether this worker is available for pushing a notification to a
#             device.
#
#             @return: ``True`` if this worker is not pushing any notification to a
#                 device; ``False`` if this worker is still processing a push
#                 notification to a device.
#             """
#             return self.is_enabled() and self.notification is None
#
#
#         def set_notification(self, notification):
#             """
#             Set the notification that this worker is requested to push to a device.
#
#             @param notification: a ``PushNotification`` instance corresponding
#                 to a notification to push to a device as indicated in this instance.
#             """
#             if self.notification is not None:
#                 raise Exception('This worker is still sending a push notification notification.')
#
#             self.notification = notification
#
#
#     def __init__(self, worker_pool_size=DEFAULT_WORKER_POOL_SIZE):
#         """
#         Build ``PushNotificationController`` instance.
#
#         @param worker_pool_size: number of workers to instantiate.
#         """
#         super(PushNotificationController, self).__init__()
#
#         self.worker_pool = [ PushNotificationController.__PushNotificationAgent__()
#                 for i in range(worker_pool_size) ]
#
#         for worker in self.worker_pool:
#             worker.start()
#
#
#     @staticmethod
#     def add_notifications(notifications):
#         """
#         Add the given notifications to be pushed to the specified devices.
#
#         @param notifications: a list of notification notification instances to push to
#             registered devices.
#         """
#         with PushNotificationController.SEMAPHORE_NOTIFICATIONS:
#             PushNotificationController.notifications.update(notifications)
#
#
#     def do(self):
#         with PushNotificationController.SEMAPHORE_NOTIFICATIONS:
#             notification_count = len(PushNotificationController.notifications)
#             if notification_count == 0:
#                 return False
#
#             available_agents = [ agent for agent in self.worker_pool if agent.is_available() ]
#
#             for i in range(min(len(available_agents), notification_count)):
#                 notification = PushNotificationController.notifications[0]
#                 if notification.schedule_time and notification.schedule_time > ISO8601DateTime.now():
#                     return i > 0
#
#                 available_agents[i].set_notification(PushNotificationController.notifications.pop())
#
#         return True
#
#
#     def shutdown(self):
#         for worker in self.worker_pool:
#             worker.shutdown()
#
#         super(PushNotificationController, self).shutdown()
#
#
# def fetch_notifications(rdbms_connection_properties,
#         batch_size=PUSH_NOTIFICATION_BATCH_SIZE):
#     with RdbmsConnection.acquire_connection(rdbms_connection_properties, auto_commit=True) as connection:
#
#         # Retrieve a list of notifications that need to be pushed to one or more
#         # devices, sorted by their schedule time in ascending order.
#         cursor = connection.execute("""
#             SELECT notification_id,
#                    notification_type,
#                    recipient_id,
#                    device_token,
#                    device_platform,
#                    sender_id,
#                    payload,
#                    utc_offset,
#                    schedule_time,
#                    expiration_time,
#                    use_local_time,
#                    notification.app_id,
#                    notification.creation_time
#                 FROM notification,
#                      notification_device
#                 WHERE notification.object_status = %(OBJECT_STATUS_ENABLED)s
#                   AND notification_mode = %(NOTIFICATION_MODE_PUSH)s
#                   AND (notification.recipient_id = notification_device.account_id::text OR
#                        notification.recipient_id = notification_device.device_id)
#                   -- AND notification.package = notification_device.package
#                   AND notification_device.object_status = %(OBJECT_STATUS_ENABLED)s
#                   AND (schedule_time IS NULL OR
#                        (NOT use_local_time AND schedule_time <= current_timestamp) OR
#                        (use_local_time AND schedule_time >= current_timestamp - '12 hours'::interval AND schedule_time < current_timestamp + '12 hours'::interval))
#                   AND is_broadcast = false -- DOES NOT SUPPORT BROADCASTING YET!
#                 ORDER BY schedule_time ASC NULLS FIRST,
#                          notification.creation_time ASC
#                 LIMIT %(limit)s
#                 FOR UPDATE""",
#             { 'NOTIFICATION_MODE_PUSH': NotificationService.NotificationMode.push,
#               'OBJECT_STATUS_ENABLED': OBJECT_STATUS_ENABLED,
#               'limit': batch_size})
#
#         notifications = [ row.get_object({
#                 'app_id': cast.string_to_uuid,
#                 'creation_time': cast.string_to_timestamp,
#                 'expiration_time': cast.string_to_timestamp,
#                 'notification_id': cast.string_to_uuid,
#                 'schedule_time': cast.string_to_timestamp })
#             for row in cursor.fetch_all() ]
#
#         if notifications:
#             connection.execute("""
#                 UPDATE notification
#                   SET object_status = %(OBJECT_STATUS_PENDING)s
#                   WHERE notification_id IN (%[notification_ids]s)""",
#                 { 'OBJECT_STATUS_PENDING': OBJECT_STATUS_PENDING,
#                   'notification_ids': set([ notification.notification_id for notification in notifications ]) })
#
#             # Convert the schedule and the expiration times to UTC if they were
#             # given in local time to the device.
#             #
#             # @todo: broadcast is not yet supported. There will be then one
#             #     notification per device, each device has its own time zone.
#             for notification in notifications:
#                 if notification.schedule_time and notification.use_local_time:
#                     notifications.schedule_time -= datetime.timedelta(hours=notification.utc_offset)
#
#                 if notification.expiration_time and notification.use_local_time:
#                     notification.expiration_time =- datetime.timedelta(hours=notification.utc_offset)
#
#                 notification.time_to_live = notification.expiration_time and \
#                         (notification.expiration_time - ISO8601DateTime.now()).total_seconds()
#
#         # Filter out notifications that have already expired.
#         return [ notification for notification in notifications
#                 if notification.time_to_live is None or notification.time_to_live > 0 ]
#
#
# def main(worker_pool_size):
#     push_notification_controller = PushNotificationController(worker_pool_size=worker_pool_size)
#     push_notification_controller.start()
#
#     try:
#         while True:
#             notifications = fetch_notifications(settings.RDBMS_CONNECTION_PROPERTIES,
#                     batch_size=PUSH_NOTIFICATION_BATCH_SIZE)
#             if len(notifications) > 0:
#                 push_notification_controller.add_notifications(notifications)
#
#             time.sleep(1)
#
#     except KeyboardInterrupt:
#         pass
#
#     except Exception, exception:
#         print traceback.format_exc()
#
#     finally:
#         push_notification_controller.shutdown()
#
#     print
#
#
# if __name__ == "__main__":
#     print "Push Notification Agent"
#     print "Copyright (C) 2013 Majormode.  All rights reserved."
#
#     parser = argparse.ArgumentParser(description='Background task used to push notifications to devices')
#     parser.add_argument('--workers', required=False, default=DEFAULT_WORKER_POOL_SIZE)
#     arguments = parser.parse_args()
#
#     main(arguments.workers)



#!/usr/bin/env python
#
# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import collections
import logging
import os
import sys

from majormode.perseus.constant.stage import EnvironmentStage
from majormode.perseus.utils import env
import psutil


# Build the default logging formatter.
DEFAULT_LOGGING_FORMATTER = logging.Formatter("%(asctime)s [%(levelname)s] %(message)s")


def __get_address_from_environment():
    """
    Return the host address of the API server instances

    The function retrieve the address defined in the environment variable
    `API_SERVER_HOSTNAME`, or defaults to the address defined in the
     constant `DEFAULT_API_SERVER_HOSTNAME` of the settings file.


    :return: The host address of the API server instances


    :raise Exception: If no address has been defined in the environment
        variable `API_SERVER_HOSTNAME`, nor in the constant
        `DEFAULT_API_SERVER_HOSTNAME` of the settings file.
    """
    import settings  # Settings of the inheriting API server application.
    address = env.getenv('API_SERVER_HOSTNAME', default_value=getattr(settings, 'DEFAULT_API_SERVER_HOSTNAME'))
    if address is None:
        raise Exception('No host address (or interface) has been defined in the environment or in the settings file')
    return address


def __get_console_handler(logging_formatter=DEFAULT_LOGGING_FORMATTER):
    """
    Return a logging handler that sends logging output to the system's
    standard output.


    :param logging_formatter: An object `Formatter` to set for this handler.


    :return: An instance of the `StreamHandler` class.
    """
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(logging_formatter)
    return console_handler


def __get_current_process_name():
    """
    Return the name of the current process


    :return: The name of the current process.
    """
    return psutil.Process().name()


def __get_opened_port_processes():
    """
    Return the list of processes that have opened TCP or UDP ports.


    :return: A dictionary where the key corresponds to a port number and
        the value corresponds to a list of `psutil.Process` that have
        opened this port.
    """
    opened_port_processes = collections.defaultdict(list)

    for process in psutil.process_iter():
        try:
            for connection in process.connections():
                if connection.status == psutil.CONN_LISTEN:
                    opened_port_processes[connection.laddr.port].append(process)
        except (psutil.AccessDenied, psutil.NoSuchProcess):
            pass

    return opened_port_processes


def __get_ports(ports=None):
    """
    Return the list of the ports of the API server instances.


    :param ports: A list of Internet Protocol port numbers which server
        application instances will listen at.  If more than one port is
        specified, the function will fork as many processes to listen at
        these additional ports.

        The argument `ports` can be an integer, a string representing a list
        of ports separated with comma, a list of integers or strings.

        If the argument `ports` is null or empty, the function retrieves the
        list of ports from the environment variable `API_SERVER_PORTS` if
        defined, otherwise the default variable `DEFAULT_API_SERVER_PORT`
        defined in the file `settings.py` of the API server project.


    :return: A list of integers corresponding to the ports of the API
        server instances.


    :raise TypeError: If the argument `ports` is not an integer, a string,
        or a list of integers/strings.

    :raise ValueError: If no ports are specified.
    """
    if ports is None:
        ports = __get_ports_from_environment()

    if isinstance(ports, int):  # integer
        ports = [ports]
    elif isinstance(ports, str):  # comma-separated values
        ports = [int(port) for port in ports.split(',')]
    elif isinstance(ports, (list, tuple, set)):
        ports = [int(port) for port in ports]
    else:
        raise TypeError("A port or a list of ports is expected")

    if len(ports) == 0:
        raise ValueError("No port number has been specified")

    return ports


def __get_ports_from_environment():
    """
    Return the ports of the API server instances

    The function retrieve the ports defined in the environment variable
    `API_SERVER_PORTS`, or defaults to the port defined in the constant
    `DEFAULT_API_SERVER_PORT` of the settings file.


    :return: A list of integers.


    :raise Exception: If no port has been defined in the environment
        variable `API_SERVER_PORTS`, nor in the constant
        `DEFAULT_API_SERVER_PORT` of the settings file.
    """
    import settings

    ports = env.getenv(
        'API_SERVER_PORTS',
        is_required=False,
        data_type=env.DataType.list,
        item_data_type=env.DataType.integer)

    if not ports:
        port = getattr(settings, 'DEFAULT_API_SERVER_PORT')
        if port is None:
            raise Exception('No port has been defined in the environment or in the settings file')
        ports = [port]

    return ports


def __load_environment_variables():
    """
    Load the environment variables from the .env file located in the
    current directory.
    """
    env_file_path_name = os.path.join(os.getcwd(), '.env')
    print(f'Loading environment file {env_file_path_name}')  # Logger not setup yet
    return env.loadenv(env_file_path_name)


def __setup_logger(
        logging_formatter=DEFAULT_LOGGING_FORMATTER,
        logging_level=logging.INFO,
        logger_name=None):
    """
    Setup a logging handler that sends logging output to the system's
    standard output.


    :param logging_formatter: An object `Formatter` to set for this handler.

    :param logger_name: Name of the logger to add the logging handler to.
        If `logger_name` is `None`, the function attaches the logging
        handler to the root logger of the hierarchy.

    :param logging_level: The threshold for the logger to `level`.  Logging
        messages which are less severe than `level` will be ignored;
        logging messages which have severity level or higher will be
        emitted by whichever handler or handlers service this logger,
        unless a handler’s level has been set to a higher severity level
        than `level`.


    :return: An object `Logger`.
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging_level)
    logger.addHandler(__get_console_handler(logging_formatter=logging_formatter))
    logger.propagate = False
    return logger


def reload(ports=None):
    raise NotImplementedError


def start(ports=None, address=None):
    """
    Run the server application on the specified port(s).


    :param ports: A list of Internet Protocol port numbers which server
        application instances will listen at.  If more than one port is
        specified, the function will fork as many processes to listen at
        these additional ports.

        The argument `ports` can be an integer, a string representing a list
        of ports separated with comma, a list of integers or strings.

        If the argument `ports` is null or empty, the function retrieves the
        list of ports from the environment variable `API_SERVER_PORTS` if
        defined, otherwise the default variable `DEFAULT_API_SERVER_PORT`
        defined in the file `settings.py` of the API server project.
        
    :param address: Address to bound the listening socket to.  Address may
        be either an IP address or hostname.  If it’s a hostname, the
        server will listen on all IP addresses associated with the name.
        Address may be an empty string or `None` to listen on all
        available interfaces.


    :raise Exception: If some of the specified ports are in use.

    :raise TypeError: If the argument `ports` is not an integer, a string,
        or a list of integers/strings.

    :raise ValueError: If no ports are specified.
    """
    __load_environment_variables()
    import settings  # Settings of the inheriting API server application.

    # Retrieve the host address and port of the API server instances.
    if address is None:
        address = __get_address_from_environment()

    ports = __get_ports(ports=ports)

    # Retrieve the logger with the specified name or, if name is `None`,
    # the logger which is the root logger of the hierarchy.
    logger_name = getattr(settings, 'LOGGER_NAME')
    logging_formatter = getattr(settings, 'LOGGING_FORMATTER')
    logging_level = getattr(settings, 'LOGGING_LEVEL')
    logger = __setup_logger(
        logging_formatter=logging_formatter,
        logging_level=logging_level,
        logger_name=logger_name)

    # Check whether some specified port numbers are already used by other
    # processes.
    opened_ports = __get_opened_port_processes()
    unavailable_ports = [port for port in ports if port in opened_ports]

    if len(unavailable_ports) > 0:
        unavailable_ports_str = ', '.join([str(port) for port in unavailable_ports])
        raise Exception(f"The following ports are already in use: {unavailable_ports_str}")

    # Fork the process to run as many instances as the number of ports
    # passed to this function, including this instance.
    from majormode.perseus.bootstrap import tornado_handler

    environment_stage = getattr(settings, 'ENVIRONMENT_STAGE', EnvironmentStage.dev)
    logger.info(f"Environment stage {environment_stage}")

    for port in ports[1:]:
        pid = os.fork()
        if pid == 0:  # Child process
            logger.info(f"Boot the API server instance ({address}:{port})")
            tornado_handler.boot(port, settings.APP_PATH, address=address)

    logger.info(f"Boot the API server instance ({address}:{ports[0]})")
    tornado_handler.boot(ports[0], settings.APP_PATH, address=address)


def stop(ports):
    """
    Terminate the current processes running that are listening TCP and
    UPD connections on the specified ports.


    :param ports: A list of Internet Protocol port numbers which server
        application instances will listen at.  If more than one port is
        specified, the function will fork as many processes to listen at
        these additional ports.

        The argument `ports` can be an integer, a string representing a list
        of ports separated with comma, a list of integers or strings.

        If the argument `ports` is null or empty, the function retrieves the
        list of ports from the environment variable `API_SERVER_PORTS` if
        defined, otherwise the default variable `DEFAULT_API_SERVER_PORT`
        defined in the file `settings.py` of the API server project.


    :return: A list of tuple `(psutil.Process, port)` of the processes
        that the function has terminated.

    :raise TypeError: If the argument `ports` is not an integer, a string,
        or a list of integers/strings.

    :raise ValueError: If no ports are specified.
    """
    __load_environment_variables()
    import settings  # Settings of the inheriting API server application.

    ports = __get_ports(ports=ports)

    # Retrieve the logger with the specified name or, if name is `None`,
    # the logger which is the root logger of the hierarchy.
    logger_name = getattr(settings, 'LOGGER_NAME')
    logging_formatter = getattr(settings, 'LOGGING_FORMATTER')
    logging_level = getattr(settings, 'LOGGING_LEVEL')
    logger = __setup_logger(
        logging_formatter=logging_formatter,
        logging_level=logging_level,
        logger_name=logger_name)

    # Search for Python processes listening at the specified ports and
    # terminate them.
    opened_port_processes = __get_opened_port_processes()
    terminated_processes = []

    for port in ports:
        for process in opened_port_processes[port]:
            try:
                logger.info(f"Process '{process.name()}' is listening at port {port}")
                if process.name() == __get_current_process_name():
                    logger.info(f"Stop the API server instance ({port})...")
                    process.terminate()
                    process.wait(5000)
                    terminated_processes.append((process, port))
            except (psutil.AccessDenied, psutil.NoSuchProcess):
                pass

    return terminated_processes



# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with
# the terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE
# SHALL NOT BE LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE
# AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
# DERIVATIVES.
#
# @version $Revision$



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with
# the terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE
# SHALL NOT BE LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE
# AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
# DERIVATIVES.
#
# @version $Revision: 274 $

from __future__ import with_statement
from xml.dom import minidom

import contextlib
import json
import jsonpickle
import sys
import time
import urllib2

class Area(object):
    AREA_TYPE_CITY = 'city'
    AREA_TYPE_DISTRICT = 'district'
    AREA_TYPE_WARD = 'ward'

    def __init__(self, name, boundary):
        names = [ n.strip() for n in name.split('/') ]
        if len(names) == 1:
            self.type = Area.AREA_TYPE_CITY
            (self.city, ) = names
        elif len(names) == 2:
            self.type = Area.AREA_TYPE_DISTRICT
            (self.city, self.district) = names
        elif len(names) == 3:
            self.type = Area.AREA_TYPE_WARD
            (self.city, self.district, self.ward) = names
        else:
            raise ValueError('The name of this area is of an unexpected composition')

        self.boundary = boundary.replace('*#', ' ').replace('*', ',').replace(' ', ',')

discovered_areas = dict()

def parse_area_data(url):
    # {s1:'21.3852558,105.801338,...',s2:'TP.Hà Nội',s3:'28'}]}

    for __attempt__ in range(3):
        try:
            with contextlib.closing(urllib2.urlopen(url)) as response:
                data = response.read()

                start_index = data.find("s1:'")
                if start_index == -1:
                    return None

                end_index = data.find("',s2:'", start_index + 4)
                boundary = data[start_index + 4:end_index]

                start_index = data.find("s2:'")
                end_index = data.find("',s3:", start_index + 4)
                name = data[start_index + 4:end_index]

                area = Area(name, boundary)
                if area.type == Area.AREA_TYPE_CITY:
                    sys.stderr.write(area.city)
                elif area.type == Area.AREA_TYPE_DISTRICT:
                    sys.stderr.write('\t' + area.district)
                else:
                    sys.stderr.write('\t\t' + area.ward)
                sys.stderr.write('\n')
                sys.stderr.flush()

                return area

        except urllib2.URLError, exception:
            sys.stderr.write(str(exception))
            time.sleep(2)


def parse_area(url):
    # <div class="NameCity">
    #  <a href="http://muabannhadat.com.vn/ban-do/ban-do-Quan-Ba-Dinh-TP-Ha-Noi-q241-t28/"
    #     class="LinkCity">
    #   Bản đồ Q.Ba Đình
    #  </a>
    # </div>
    if discovered_areas.has_key(url):
        return

    for _attempt_ in range(3):
        try:
            with contextlib.closing(urllib2.urlopen(url)) as response:
                data = response.read()

                start_index = data.find('http://muabannhadat.com.vn/nvc/')
                if start_index > 0:
                    end_index = data.find('"', start_index)
                    area = parse_area_data(data[start_index:end_index])
                    if area is not None:
                        discovered_areas[url] = area

                start_index = data.find('<div class="CityManila">')
                end_index = data.find('\t<div class="clr_both">')
                xmldoc = minidom.parseString(data[start_index:end_index]).documentElement
                for divnode in xmldoc.getElementsByTagName('div'):
                    if divnode.getAttribute('class') != 'NameCity':
                        raise Exception("Unexpected HTML page format: doesn't contain div class of NameCity")
                    if len(divnode.childNodes) != 1:
                        raise Exception('Unexpected number of URL links to a geographic area')
                    anode = divnode.childNodes[0]
                    if anode.getAttribute('class') != 'LinkCity':
                        raise Exception('Unexpected class for a HTML reference of a geographic area')
                    parse_area(anode.getAttribute('href'))
            break
        except urllib2.URLError, exception:
            sys.stderr.write(str(exception))
            time.sleep(2)

if __name__ == "__main__":
    parse_area('http://muabannhadat.com.vn/ban-do/ban-do-viet-nam/')
    sys.stdout.write(json.dumps(jsonpickle.Pickler(unpicklable=False).flatten(discovered_areas.values()), indent=2))
    sys.stdout.flush()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with
# the terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE
# SHALL NOT BE LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE
# AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
# DERIVATIVES.
#
# @version $Revision: 344 $

from __future__ import with_statement

from majormode.perseus.model.version import Version
from majormode.perseus.service.version.version_service import VersionService
from majormode.perseus.unittest.base_unittest import BaseApiKeyTestCase

import random
import settings
import sys
import unittest

class VersionServiceTestCase(BaseApiKeyTestCase):
    @staticmethod
    def __version_generator__():
        while True:
            major = random.randint(0, sys.maxint)
            minor = random.randint(0, sys.maxint)
            patch = random.randint(0, sys.maxint)
            yield Version('%s.%s.%s' % (major, minor, patch))

    def setup(self):
        super(VersionServiceTestCase, self).__init__()
        random.seed()

    def test_version_wellformed(self):
        """
        Test the well-formed string representations of version.
        """
        Version('0.0.0')
        Version('1.0.0')
        Version('1.2.0')
        Version('1.2.3')
        _ = [ self.__version_generator__().next() for _ in xrange(100) ]

    def test_versions_malformed(self):
        """
        Test the malformed string representations of version.
        """
        self.assertRaises(Version.InvalidVersionFormatException, Version, 'a.b.c', True)
        self.assertRaises(Version.InvalidVersionFormatException, Version, '1.b.c', True)
        self.assertRaises(Version.InvalidVersionFormatException, Version, '1.2.c', True)

        self.assertRaises(Version.InvalidVersionFormatException, Version, '-1.-1.-1', True)
        self.assertRaises(Version.InvalidVersionFormatException, Version, '-1.-1.3', True)
        self.assertRaises(Version.InvalidVersionFormatException, Version, '-1.2.3', True)

        self.assertRaises(Version.InvalidVersionFormatException, Version, '1a.2.3', True)
        self.assertRaises(Version.InvalidVersionFormatException, Version, '1.2b.3', True)
        self.assertRaises(Version.InvalidVersionFormatException, Version, '1.2.3c', True)

    def test_version_comparisons(self):
        """
        Test common comparison operations of version objects.
        """
        self.assertEqual(Version('0.0.0'), Version('0.0.0'))
        self.assertEqual(Version('1.0.0'), Version('1.0.0'))
        self.assertEqual(Version('1.2.0'), Version('1.2.0'))
        self.assertEqual(Version('1.2.3'), Version('1.2.3'))

        self.assertGreater(Version('1.2.3'), Version('1.2.2'))
        self.assertLess(Version('1.2.2'), Version('1.2.3'))
        self.assertGreater(Version('1.2.3'), Version('1.1.4'))
        self.assertLess(Version('1.1.4'), Version('1.2.3'))
        self.assertGreater(Version('2.2.3'), Version('2.1.4'))

        generator = self.__version_generator__()
        for (a, b) in [ (generator.next(), generator.next()) for _ in xrange(100) ]:
            if a.major > b.major:
                self.assertGreater(a, b)
            elif a.major < b.major:
                self.assertLess(a, b)
            elif a.minor > b.minor:
                self.assertGreater(a, b)
            elif a.minor < b.minor:
                self.assertLess(a, b)
            elif a.patch > b.patch:
                self.assertGreater(a, b)
            elif a.patch < b.patch:
                self.assertLess(a, b)
            else:
                self.assertEqual(a, b)

    def test_api_version(self):
        """
        Get the version of the platform API currently deployed.
        """
        self.assertEqual(VersionService().get_version(self.unittest_app_id),
                         Version(settings.API_VERSION))

def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.makeSuite(VersionServiceTestCase))
    return test_suite

if __name__ == '__main__':
    unittest.main()
